#!/usr/bin/python

# Copyright (c) 2017, Giovanni Sciortino (@giovannisciortino)
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

DOCUMENTATION = r"""
module: rhsm_repository
short_description: Manage RHSM repositories using the subscription-manager command
description:
  - Manage (Enable/Disable) RHSM repositories to the Red Hat Subscription Management entitlement platform using the C(subscription-manager)
    command.
author: Giovanni Sciortino (@giovannisciortino)
notes:
  - In order to manage RHSM repositories the system must be already registered to RHSM manually or using the Ansible M(community.general.redhat_subscription)
    module.
  - It is possible to interact with C(subscription-manager) only as root, so root permissions are required to successfully
    run this module.
requirements:
  - subscription-manager
extends_documentation_fragment:
  - community.general._attributes
attributes:
  check_mode:
    support: full
  diff_mode:
    support: full
options:
  state:
    description:
      - If state is equal to present or disabled, indicates the desired repository state.
      - In community.general 10.0.0 the states V(present) and V(absent) have been removed. Please use V(enabled) and V(disabled)
        instead.
    choices: [enabled, disabled]
    default: "enabled"
    type: str
  name:
    description:
      - The ID of repositories to enable.
      - To operate on several repositories this can accept a comma separated list or a YAML list.
    required: true
    type: list
    elements: str
  purge:
    description:
      - Disable all currently enabled repositories that are not not specified in O(name). Only set this to V(true) if passing
        in a list of repositories to the O(name) field. Using this with C(loop) is likely not to have the desired result.
    type: bool
    default: false
"""

EXAMPLES = r"""
- name: Enable a RHSM repository
  community.general.rhsm_repository:
    name: rhel-7-server-rpms

- name: Disable all RHSM repositories
  community.general.rhsm_repository:
    name: '*'
    state: disabled

- name: Enable all repositories starting with rhel-6-server
  community.general.rhsm_repository:
    name: rhel-6-server*
    state: enabled

- name: Disable all repositories except rhel-7-server-rpms
  community.general.rhsm_repository:
    name: rhel-7-server-rpms
    purge: true
"""

RETURN = r"""
repositories:
  description:
    - The list of RHSM repositories with their states.
    - When this module is used to change the repository states, this list contains the updated states after the changes.
  returned: success
  type: list
"""

import os
from copy import deepcopy
from fnmatch import fnmatch

from ansible.module_utils.basic import AnsibleModule


class Rhsm:
    def __init__(self, module):
        self.module = module
        self.rhsm_bin = self.module.get_bin_path("subscription-manager", required=True)
        self.rhsm_kwargs = {
            "environ_update": dict(LANGUAGE="C", LC_ALL="C"),
            "expand_user_and_vars": False,
            "use_unsafe_shell": False,
        }

    def run_repos(self, arguments):
        """
        Execute `subscription-manager repos` with arguments and manage common errors
        """
        rc, out, err = self.module.run_command([self.rhsm_bin, "repos"] + arguments, **self.rhsm_kwargs)

        if rc == 0 and out == "This system has no repositories available through subscriptions.\n":
            self.module.fail_json(msg="This system has no repositories available through subscriptions")
        elif rc == 1:
            self.module.fail_json(msg=f"subscription-manager failed with the following error: {err}")
        else:
            return rc, out, err

    def list_repositories(self):
        """
        Generate RHSM repository list and return a list of dict
        """
        rc, out, err = self.run_repos(["--list"])

        repo_id = ""
        repo_name = ""
        repo_url = ""
        repo_enabled = ""

        repo_result = []
        for line in out.splitlines():
            # ignore lines that are:
            # - empty
            # - "+---------[...]" -- i.e. header
            # - "    Available Repositories [...]" -- i.e. header
            if line == "" or line[0] == "+" or line[0] == " ":
                continue

            if line.startswith("Repo ID: "):
                repo_id = line[9:].lstrip()
                continue

            if line.startswith("Repo Name: "):
                repo_name = line[11:].lstrip()
                continue

            if line.startswith("Repo URL: "):
                repo_url = line[10:].lstrip()
                continue

            if line.startswith("Enabled: "):
                repo_enabled = line[9:].lstrip()

                repo = {
                    "id": repo_id,
                    "name": repo_name,
                    "url": repo_url,
                    "enabled": True if repo_enabled == "1" else False,
                }

                repo_result.append(repo)

        return repo_result


def repository_modify(module, rhsm, state, name, purge=False):
    name = set(name)
    current_repo_list = rhsm.list_repositories()
    updated_repo_list = deepcopy(current_repo_list)
    matched_existing_repo = {}
    for repoid in name:
        matched_existing_repo[repoid] = []
        for idx, repo in enumerate(current_repo_list):
            if fnmatch(repo["id"], repoid):
                matched_existing_repo[repoid].append(repo)
                # Update current_repo_list to return it as result variable
                updated_repo_list[idx]["enabled"] = True if state == "enabled" else False

    changed = False
    results = []
    diff_before = ""
    diff_after = ""
    rhsm_arguments = []

    for repoid in matched_existing_repo:
        if len(matched_existing_repo[repoid]) == 0:
            results.append(f"{repoid} is not a valid repository ID")
            module.fail_json(results=results, msg=f"{repoid} is not a valid repository ID")
        for repo in matched_existing_repo[repoid]:
            if state in ["disabled", "absent"]:
                if repo["enabled"]:
                    changed = True
                    diff_before += f"Repository '{repo['id']}' is enabled for this system\n"
                    diff_after += f"Repository '{repo['id']}' is disabled for this system\n"
                results.append(f"Repository '{repo['id']}' is disabled for this system")
                rhsm_arguments += ["--disable", repo["id"]]
            elif state in ["enabled", "present"]:
                if not repo["enabled"]:
                    changed = True
                    diff_before += f"Repository '{repo['id']}' is disabled for this system\n"
                    diff_after += f"Repository '{repo['id']}' is enabled for this system\n"
                results.append(f"Repository '{repo['id']}' is enabled for this system")
                rhsm_arguments += ["--enable", repo["id"]]

    # Disable all enabled repos on the system that are not in the task and not
    # marked as disabled by the task
    if purge:
        enabled_repo_ids = {repo["id"] for repo in updated_repo_list if repo["enabled"]}
        matched_repoids_set = set(matched_existing_repo.keys())
        difference = enabled_repo_ids.difference(matched_repoids_set)
        if len(difference) > 0:
            for repoid in difference:
                changed = True
                diff_before.join(f"Repository '{repoid}'' is enabled for this system\n")
                diff_after.join(f"Repository '{repoid}' is disabled for this system\n")
                results.append(f"Repository '{repoid}' is disabled for this system")
                rhsm_arguments.extend(["--disable", repoid])
            for updated_repo in updated_repo_list:
                if updated_repo["id"] in difference:
                    updated_repo["enabled"] = False

    diff = {
        "before": diff_before,
        "after": diff_after,
        "before_header": "RHSM repositories",
        "after_header": "RHSM repositories",
    }

    if not module.check_mode and changed:
        rc, out, err = rhsm.run_repos(rhsm_arguments)
        results = out.splitlines()
    module.exit_json(results=results, changed=changed, repositories=updated_repo_list, diff=diff)


def main():
    module = AnsibleModule(
        argument_spec=dict(
            name=dict(type="list", elements="str", required=True),
            state=dict(choices=["enabled", "disabled"], default="enabled"),
            purge=dict(type="bool", default=False),
        ),
        supports_check_mode=True,
    )

    if os.getuid() != 0:
        module.fail_json(msg="Interacting with subscription-manager requires root permissions ('become: true')")

    rhsm = Rhsm(module)

    name = module.params["name"]
    state = module.params["state"]
    purge = module.params["purge"]

    repository_modify(module, rhsm, state, name, purge)


if __name__ == "__main__":
    main()



#!/usr/bin/python

# Copyright (c) 2016, Hugh Ma <Hugh.Ma@flextronics.com>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

DOCUMENTATION = r"""
module: stacki_host
short_description: Add or remove host to stacki front-end
description:
  - Use this module to add or remove hosts to a stacki front-end using API.
  - Information on stacki can be found at U(https://github.com/StackIQ/stacki).
extends_documentation_fragment:
  - community.general._attributes
attributes:
  check_mode:
    support: none
  diff_mode:
    support: none
options:
  name:
    description:
      - Name of the host to be added to Stacki.
    required: true
    type: str
  stacki_user:
    description:
      - Username for authenticating with Stacki API, but if not specified, the environment variable E(stacki_user) is used
        instead.
    required: true
    type: str
  stacki_password:
    description:
      - Password for authenticating with Stacki API, but if not specified, the environment variable E(stacki_password) is
        used instead.
    required: true
    type: str
  stacki_endpoint:
    description:
      - URL for the Stacki API Endpoint.
    required: true
    type: str
  prim_intf_mac:
    description:
      - MAC Address for the primary PXE boot network interface.
      - Currently not used by the module.
    type: str
  prim_intf_ip:
    description:
      - IP Address for the primary network interface.
      - Currently not used by the module.
    type: str
  prim_intf:
    description:
      - Name of the primary network interface.
      - Currently not used by the module.
    type: str
  force_install:
    description:
      - Set value to V(true) to force node into install state if it already exists in stacki.
    type: bool
    default: false
  state:
    description:
      - Set value to the desired state for the specified host.
    type: str
    choices: [absent, present]
    default: present
  appliance:
    description:
      - Appliance to be used in host creation.
      - Required if O(state=present) and host does not yet exist.
    type: str
    default: backend
  rack:
    description:
      - Rack to be used in host creation.
      - Required if O(state=present) and host does not yet exist.
    type: int
    default: 0
  rank:
    description:
      - Rank to be used in host creation.
      - In Stacki terminology, the rank is the position of the machine in a rack.
      - Required if O(state=present) and host does not yet exist.
    type: int
    default: 0
  network:
    description:
      - Network to be configured in the host.
      - Currently not used by the module.
    type: str
    default: private
author:
  - Hugh Ma (@bbyhuy) <Hugh.Ma@flextronics.com>
"""

EXAMPLES = r"""
- name: Add a host named test-1
  community.general.stacki_host:
    name: test-1
    stacki_user: usr
    stacki_password: pwd
    stacki_endpoint: url
    prim_intf_mac: mac_addr
    prim_intf_ip: x.x.x.x
    prim_intf: eth0

- name: Remove a host named test-1
  community.general.stacki_host:
    name: test-1
    stacki_user: usr
    stacki_password: pwd
    stacki_endpoint: url
    state: absent
"""


import json
from urllib.parse import urlencode

from ansible.module_utils.basic import AnsibleModule, env_fallback
from ansible.module_utils.urls import fetch_url


class StackiHost:
    def __init__(self, module):
        self.module = module
        self.hostname = module.params["name"]
        self.rack = module.params["rack"]
        self.rank = module.params["rank"]
        self.appliance = module.params["appliance"]
        self.prim_intf = module.params["prim_intf"]
        self.prim_intf_ip = module.params["prim_intf_ip"]
        self.network = module.params["network"]
        self.prim_intf_mac = module.params["prim_intf_mac"]
        self.endpoint = module.params["stacki_endpoint"]

        auth_creds = {"USERNAME": module.params["stacki_user"], "PASSWORD": module.params["stacki_password"]}

        # Get Initial CSRF
        cred_a = self.do_request(self.endpoint, method="GET")
        cookie_a = cred_a.headers.get("Set-Cookie").split(";")
        init_csrftoken = None
        for c in cookie_a:
            if "csrftoken" in c:
                init_csrftoken = c.replace("csrftoken=", "")
                init_csrftoken = init_csrftoken.rstrip("\r\n")
                break

        # Make Header Dictionary with initial CSRF
        header = {
            "csrftoken": init_csrftoken,
            "X-CSRFToken": init_csrftoken,
            "Content-type": "application/x-www-form-urlencoded",
            "Cookie": cred_a.headers.get("Set-Cookie"),
        }

        # Endpoint to get final authentication header
        login_endpoint = f"{self.endpoint}/login"

        # Get Final CSRF and Session ID
        login_req = self.do_request(login_endpoint, headers=header, payload=urlencode(auth_creds), method="POST")

        cookie_f = login_req.headers.get("Set-Cookie").split(";")
        csrftoken = None
        for f in cookie_f:
            if "csrftoken" in f:
                csrftoken = f.replace("csrftoken=", "")
            if "sessionid" in f:
                sessionid = c.split("sessionid=", 1)[-1]
                sessionid = sessionid.rstrip("\r\n")

        self.header = {
            "csrftoken": csrftoken,
            "X-CSRFToken": csrftoken,
            "sessionid": sessionid,
            "Content-type": "application/json",
            "Cookie": login_req.headers.get("Set-Cookie"),
        }

    def do_request(self, url, payload=None, headers=None, method=None):
        res, info = fetch_url(self.module, url, data=payload, headers=headers, method=method)

        if info["status"] != 200:
            self.module.fail_json(changed=False, msg=info["msg"])

        return res

    def stack_check_host(self):
        res = self.do_request(
            self.endpoint, payload=json.dumps({"cmd": "list host"}), headers=self.header, method="POST"
        )
        return self.hostname in res.read()

    def stack_sync(self):
        self.do_request(self.endpoint, payload=json.dumps({"cmd": "sync config"}), headers=self.header, method="POST")
        self.do_request(
            self.endpoint, payload=json.dumps({"cmd": "sync host config"}), headers=self.header, method="POST"
        )

    def stack_force_install(self, result):
        data = {"cmd": f"set host boot {self.hostname} action=install"}
        self.do_request(self.endpoint, payload=json.dumps(data), headers=self.header, method="POST")
        changed = True

        self.stack_sync()

        result["changed"] = changed
        result["stdout"] = "api call successful".rstrip("\r\n")

    def stack_add(self, result):
        data = dict()
        changed = False

        data["cmd"] = f"add host {self.hostname} rack={self.rack} rank={self.rank} appliance={self.appliance}"
        self.do_request(self.endpoint, payload=json.dumps(data), headers=self.header, method="POST")

        self.stack_sync()

        result["changed"] = changed
        result["stdout"] = "api call successful".rstrip("\r\n")

    def stack_remove(self, result):
        data = dict()

        data["cmd"] = f"remove host {self.hostname}"
        self.do_request(self.endpoint, payload=json.dumps(data), headers=self.header, method="POST")

        self.stack_sync()

        result["changed"] = True
        result["stdout"] = "api call successful".rstrip("\r\n")


def main():
    module = AnsibleModule(
        argument_spec=dict(
            state=dict(type="str", default="present", choices=["absent", "present"]),
            name=dict(type="str", required=True),
            rack=dict(type="int", default=0),
            rank=dict(type="int", default=0),
            appliance=dict(type="str", default="backend"),
            prim_intf=dict(type="str"),
            prim_intf_ip=dict(type="str"),
            network=dict(type="str", default="private"),
            prim_intf_mac=dict(type="str"),
            stacki_user=dict(type="str", required=True, fallback=(env_fallback, ["stacki_user"])),
            stacki_password=dict(type="str", required=True, fallback=(env_fallback, ["stacki_password"]), no_log=True),
            stacki_endpoint=dict(type="str", required=True, fallback=(env_fallback, ["stacki_endpoint"])),
            force_install=dict(type="bool", default=False),
        ),
        supports_check_mode=False,
    )

    result = {"changed": False}
    missing_params = list()

    stacki = StackiHost(module)
    host_exists = stacki.stack_check_host()

    # If state is present, but host exists, need force_install flag to put host back into install state
    if module.params["state"] == "present" and host_exists and module.params["force_install"]:
        stacki.stack_force_install(result)
    # If state is present, but host exists, and force_install and false, do nothing
    elif module.params["state"] == "present" and host_exists and not module.params["force_install"]:
        result["stdout"] = f"{module.params['name']} already exists. Set 'force_install' to true to bootstrap"
    # Otherwise, state is present, but host doesn't exists, require more params to add host
    elif module.params["state"] == "present" and not host_exists:
        for param in ["appliance", "rack", "rank", "prim_intf", "prim_intf_ip", "network", "prim_intf_mac"]:
            if not module.params[param]:
                missing_params.append(param)
        if len(missing_params) > 0:
            module.fail_json(msg=f"missing required arguments: {missing_params}")

        stacki.stack_add(result)
    # If state is absent, and host exists, lets remove it.
    elif module.params["state"] == "absent" and host_exists:
        stacki.stack_remove(result)

    module.exit_json(**result)


if __name__ == "__main__":
    main()



#!/usr/bin/python

# Copyright (c) 2016, Dag Wieers <dag@wieers.com>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

DOCUMENTATION = r"""
module: wakeonlan
short_description: Send a magic Wake-on-LAN (WoL) broadcast packet
description:
  - The C(wakeonlan) module sends magic Wake-on-LAN (WoL) broadcast packets.
extends_documentation_fragment:
  - community.general._attributes
attributes:
  check_mode:
    support: full
  diff_mode:
    support: none
options:
  mac:
    description:
      - MAC address to send Wake-on-LAN broadcast packet for.
    required: true
    type: str
  broadcast:
    description:
      - Network broadcast address to use for broadcasting magic Wake-on-LAN packet.
    default: 255.255.255.255
    type: str
  port:
    description:
      - UDP port to use for magic Wake-on-LAN packet.
    default: 7
    type: int
todo:
  - Add arping support to check whether the system is up (before and after)
  - Enable check-mode support (when we have arping support)
  - Does not have SecureOn password support
notes:
  - This module sends a magic packet, without knowing whether it worked.
  - Only works if the target system was properly configured for Wake-on-LAN (in the BIOS and/or the OS).
  - Some BIOSes have a different (configurable) Wake-on-LAN boot order (in other words, PXE first).
seealso:
  - module: community.windows.win_wakeonlan
author:
  - Dag Wieers (@dagwieers)
"""

EXAMPLES = r"""
- name: Send a magic Wake-on-LAN packet to 00:00:5E:00:53:66
  community.general.wakeonlan:
    mac: '00:00:5E:00:53:66'
    broadcast: 192.0.2.23
  delegate_to: localhost

- community.general.wakeonlan:
    mac: 00:00:5E:00:53:66
    port: 9
  delegate_to: localhost
"""

RETURN = r"""
# Default return values
"""
import socket
import struct
import traceback

from ansible.module_utils.basic import AnsibleModule


def wakeonlan(module, mac, broadcast, port):
    """Send a magic Wake-on-LAN packet."""

    mac_orig = mac

    # Remove possible separator from MAC address
    if len(mac) == 12 + 5:
        mac = mac.replace(mac[2], "")

    # If we don't end up with 12 hexadecimal characters, fail
    if len(mac) != 12:
        module.fail_json(msg=f"Incorrect MAC address length: {mac_orig}")

    # Test if it converts to an integer, otherwise fail
    try:
        int(mac, 16)
    except ValueError:
        module.fail_json(msg=f"Incorrect MAC address format: {mac_orig}")

    # Create payload for magic packet
    data = b""
    padding = f"FFFFFFFFFFFF{mac * 20}"
    for i in range(0, len(padding), 2):
        data = b"".join([data, struct.pack("B", int(padding[i : i + 2], 16))])

    # Broadcast payload to network
    sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    sock.setsockopt(socket.SOL_SOCKET, socket.SO_BROADCAST, 1)

    if not module.check_mode:
        try:
            sock.sendto(data, (broadcast, port))
        except OSError as e:
            sock.close()
            module.fail_json(msg=f"{e}", exception=traceback.format_exc())

    sock.close()


def main():
    module = AnsibleModule(
        argument_spec=dict(
            mac=dict(type="str", required=True),
            broadcast=dict(type="str", default="255.255.255.255"),
            port=dict(type="int", default=7),
        ),
        supports_check_mode=True,
    )

    mac = module.params["mac"]
    broadcast = module.params["broadcast"]
    port = module.params["port"]

    wakeonlan(module, mac, broadcast, port)

    module.exit_json(changed=True)


if __name__ == "__main__":
    main()



#!/usr/bin/python

# Copyright (c) 2016, Adam Števko <adam.stevko@gmail.com>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

DOCUMENTATION = r"""
module: zfs_facts
short_description: Gather facts about ZFS datasets
description:
  - Gather facts from ZFS dataset properties.
author: Adam Števko (@xen0l)
extends_documentation_fragment:
  - community.general._attributes
  - community.general._attributes.facts
  - community.general._attributes.facts_module
options:
  name:
    description:
      - ZFS dataset name.
    required: true
    aliases: ["ds", "dataset"]
    type: str
  recurse:
    description:
      - Specifies if properties for any children should be recursively displayed.
    type: bool
    default: false
  parsable:
    description:
      - Specifies if property values should be displayed in machine friendly format.
    type: bool
    default: false
  properties:
    description:
      - Specifies which dataset properties should be queried in comma-separated format. For more information about dataset
        properties, check zfs(1M) man page.
    default: all
    type: str
  type:
    description:
      - Specifies which datasets types to display. Multiple values have to be provided as a list or in comma-separated form.
      - Value V(all) cannot be used together with other values.
    choices: ['all', 'filesystem', 'volume', 'snapshot', 'bookmark']
    default: [all]
    type: list
    elements: str
  depth:
    description:
      - Specifies recursion depth.
    type: int
    default: 0
"""

EXAMPLES = r"""
- name: Gather facts about ZFS dataset rpool/export/home
  community.general.zfs_facts:
    dataset: rpool/export/home

- name: Report space usage on ZFS filesystems under data/home
  community.general.zfs_facts:
    name: data/home
    recurse: true
    type: filesystem

- ansible.builtin.debug:
    msg: 'ZFS dataset {{ item.name }} consumes {{ item.used }} of disk space.'
  with_items: '{{ ansible_zfs_datasets }}'
"""

RETURN = r"""
name:
  description: ZFS dataset name.
  returned: success
  type: str
  sample: rpool/var/spool
parsable:
  description: If parsable output should be provided in machine friendly format.
  returned: if O(parsable=True)
  type: bool
  sample: true
recurse:
  description: If we should recurse over ZFS dataset.
  returned: if O(recurse=True)
  type: bool
  sample: true
ansible_facts:
  description: Dictionary containing all the detailed information about the ZFS facts.
  returned: success
  type: dict
  contains:
    zfs_datasets:
      description: ZFS dataset facts.
      returned: always
      type: list
      elements: dict
      sample:
        - "aclinherit": "restricted"
          "aclmode": "discard"
          "atime": "on"
          "available": "43.8G"
          "canmount": "on"
          "casesensitivity": "sensitive"
          "checksum": "on"
          "compression": "off"
          "compressratio": "1.00x"
          "copies": "1"
          "creation": "Thu Jun 16 11:37 2016"
          "dedup": "off"
          "devices": "on"
          "exec": "on"
          "filesystem_count": "none"
          "filesystem_limit": "none"
          "logbias": "latency"
          "logicalreferenced": "18.5K"
          "logicalused": "3.45G"
          "mlslabel": "none"
          "mounted": "yes"
          "mountpoint": "/rpool"
          "name": "rpool"
          "nbmand": "off"
          "normalization": "none"
          "org.openindiana.caiman:install": "ready"
          "primarycache": "all"
          "quota": "none"
          "readonly": "off"
          "recordsize": "128K"
          "redundant_metadata": "all"
          "refcompressratio": "1.00x"
          "referenced": "29.5K"
          "refquota": "none"
          "refreservation": "none"
          "reservation": "none"
          "secondarycache": "all"
          "setuid": "on"
          "sharenfs": "off"
          "sharesmb": "off"
          "snapdir": "hidden"
          "snapshot_count": "none"
          "snapshot_limit": "none"
          "sync": "standard"
          "type": "filesystem"
          "used": "4.41G"
          "usedbychildren": "4.41G"
          "usedbydataset": "29.5K"
          "usedbyrefreservation": "0"
          "usedbysnapshots": "0"
          "utf8only": "off"
          "version": "5"
          "vscan": "off"
          "written": "29.5K"
          "xattr": "on"
          "zoned": "off"
"""

from collections import defaultdict

from ansible.module_utils.basic import AnsibleModule

SUPPORTED_TYPES = ["all", "filesystem", "volume", "snapshot", "bookmark"]


class ZFSFacts:
    def __init__(self, module):
        self.module = module

        self.name = module.params["name"]
        self.recurse = module.params["recurse"]
        self.parsable = module.params["parsable"]
        self.properties = module.params["properties"]
        self.type = module.params["type"]
        self.depth = module.params["depth"]

        self._datasets = defaultdict(dict)
        self.facts = []

    def dataset_exists(self):
        cmd = [self.module.get_bin_path("zfs"), "list", self.name]

        (rc, out, err) = self.module.run_command(cmd)

        return rc == 0

    def get_facts(self):
        cmd = [self.module.get_bin_path("zfs"), "get", "-H"]
        if self.parsable:
            cmd.append("-p")
        if self.recurse:
            cmd.append("-r")
        if self.depth != 0:
            cmd.append("-d")
            cmd.append(f"{self.depth}")
        if self.type:
            cmd.append("-t")
            cmd.append(",".join(self.type))
        cmd.extend(["-o", "name,property,value", self.properties, self.name])

        (rc, out, err) = self.module.run_command(cmd, check_rc=True)

        for line in out.splitlines():
            dataset, property, value = line.split("\t")

            self._datasets[dataset].update({property: value})

        for k, v in self._datasets.items():
            v.update({"name": k})
            self.facts.append(v)

        return {"ansible_zfs_datasets": self.facts}


def main():
    module = AnsibleModule(
        argument_spec=dict(
            name=dict(required=True, aliases=["ds", "dataset"], type="str"),
            recurse=dict(default=False, type="bool"),
            parsable=dict(default=False, type="bool"),
            properties=dict(default="all", type="str"),
            type=dict(default="all", type="list", elements="str", choices=SUPPORTED_TYPES),
            depth=dict(default=0, type="int"),
        ),
        supports_check_mode=True,
    )
    module.run_command_environ_update = {"LANGUAGE": "C", "LC_ALL": "C"}

    if "all" in module.params["type"] and len(module.params["type"]) > 1:
        module.fail_json(msg="Value 'all' for parameter 'type' is mutually exclusive with other values")

    zfs_facts = ZFSFacts(module)

    result = {}
    result["changed"] = False
    result["name"] = zfs_facts.name

    if zfs_facts.parsable:
        result["parsable"] = zfs_facts.parsable

    if zfs_facts.recurse:
        result["recurse"] = zfs_facts.recurse

    if not zfs_facts.dataset_exists():
        module.fail_json(msg=f"ZFS dataset {zfs_facts.name} does not exist!")

    result["ansible_facts"] = zfs_facts.get_facts()

    module.exit_json(**result)


if __name__ == "__main__":
    main()



#!/usr/bin/python
#
# Copyright (c) 2024, Alexander Bakanovskii <skottttt228@gmail.com>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

# Make coding more python3-ish
from __future__ import annotations

DOCUMENTATION = r"""
module: one_vnet
short_description: Manages OpenNebula virtual networks
version_added: 9.4.0
author: "Alexander Bakanovskii (@abakanovskii)"
requirements:
  - pyone
description:
  - Manages virtual networks in OpenNebula.
attributes:
  check_mode:
    support: partial
    details:
      - Note that check mode always returns RV(ignore:changed=true) for existing networks, even if the network would not actually
        change.
  diff_mode:
    support: none
options:
  id:
    description:
      - A O(id) of the network you would like to manage.
      - If not set then a new network is created with the given O(name).
    type: int
  name:
    description:
      - A O(name) of the network you would like to manage. If a network with the given name does not exist it, then is created,
        otherwise it is managed by this module.
    type: str
  template:
    description:
      - A string containing the network template contents.
    type: str
  state:
    description:
      - V(present) - state that is used to manage the network.
      - V(absent) - delete the network.
    choices: ["present", "absent"]
    default: present
    type: str

extends_documentation_fragment:
  - community.general._opennebula
  - community.general._attributes
"""

EXAMPLES = r"""
- name: Make sure the network is present by ID
  community.general.one_vnet:
    id: 0
    state: present
  register: result

- name: Make sure the network is present by name
  community.general.one_vnet:
    name: opennebula-bridge
    state: present
  register: result

- name: Create a new or update an existing network
  community.general.one_vnet:
    name: bridge-network
    template: |
      VN_MAD  = "bridge"
      BRIDGE  = "br0"
      BRIDGE_TYPE  = "linux"
      AR=[
        TYPE  = "IP4",
        IP    = 192.0.2.50,
        SIZE  = "20"
      ]
      DNS     = 192.0.2.1
      GATEWAY = 192.0.2.1

- name: Delete the network by ID
  community.general.one_vnet:
    id: 0
    state: absent
"""

RETURN = r"""
id:
  description: The network ID.
  type: int
  returned: when O(state=present)
  sample: 153
name:
  description: The network name.
  type: str
  returned: when O(state=present)
  sample: app1
template:
  description: The parsed network template.
  type: dict
  returned: when O(state=present)
  sample:
    BRIDGE: onebr.1000
    BRIDGE_TYPE: linux
    DESCRIPTION: sampletext
    PHYDEV: eth0
    SECURITY_GROUPS: 0
    VLAN_ID: 1000
    VN_MAD: 802.1Q
user_id:
  description: The network's user name.
  type: int
  returned: when O(state=present)
  sample: 1
user_name:
  description: The network's user ID.
  type: str
  returned: when O(state=present)
  sample: oneadmin
group_id:
  description: The network's group ID.
  type: int
  returned: when O(state=present)
  sample: 1
group_name:
  description: The network's group name.
  type: str
  returned: when O(state=present)
  sample: one-users
owner_id:
  description: The network's owner ID.
  type: int
  returned: when O(state=present)
  sample: 143
owner_name:
  description: The network's owner name.
  type: str
  returned: when O(state=present)
  sample: ansible-test
permissions:
  description: The network's permissions.
  type: dict
  returned: when O(state=present)
  contains:
    owner_u:
      description: The network's owner USAGE permissions.
      type: str
      sample: 1
    owner_m:
      description: The network's owner MANAGE permissions.
      type: str
      sample: 0
    owner_a:
      description: The network's owner ADMIN permissions.
      type: str
      sample: 0
    group_u:
      description: The network's group USAGE permissions.
      type: str
      sample: 0
    group_m:
      description: The network's group MANAGE permissions.
      type: str
      sample: 0
    group_a:
      description: The network's group ADMIN permissions.
      type: str
      sample: 0
    other_u:
      description: The network's other users USAGE permissions.
      type: str
      sample: 0
    other_m:
      description: The network's other users MANAGE permissions.
      type: str
      sample: 0
    other_a:
      description: The network's other users ADMIN permissions.
      type: str
      sample: 0
  sample:
    owner_u: 1
    owner_m: 0
    owner_a: 0
    group_u: 0
    group_m: 0
    group_a: 0
    other_u: 0
    other_m: 0
    other_a: 0
clusters:
  description: The network's clusters.
  type: list
  returned: when O(state=present)
  sample: [0, 100]
bridge:
  description: The network's bridge interface.
  type: str
  returned: when O(state=present)
  sample: br0
bridge_type:
  description: The network's bridge type.
  type: str
  returned: when O(state=present)
  sample: linux
parent_network_id:
  description: The network's parent network ID.
  type: int
  returned: when O(state=present)
  sample: 1
vn_mad:
  description: The network's VN_MAD.
  type: str
  returned: when O(state=present)
  sample: bridge
phydev:
  description: The network's physical device (NIC).
  type: str
  returned: when O(state=present)
  sample: eth0
vlan_id:
  description: The network's VLAN tag.
  type: int
  returned: when O(state=present)
  sample: 1000
outer_vlan_id:
  description: The network's outer VLAN tag.
  type: int
  returned: when O(state=present)
  sample: 1000
vrouters:
  description: The network's list of virtual routers IDs.
  type: list
  returned: when O(state=present)
  sample: [0, 1]
ar_pool:
  description: The network's list of ar_pool.
  type: list
  returned: when O(state=present)
  sample:
    - ar_id: 0
      ip: 192.0.2.1
      mac: 6c:1e:46:01:cd:d1
      size: 20
      type: IP4
    - ar_id: 1
      allocated: 0
      ip: 198.51.100.1
      mac: 5d:9b:c0:9e:f6:e5
      size: 20
      type: IP4
"""


from ansible_collections.community.general.plugins.module_utils._opennebula import OpenNebulaModule


class NetworksModule(OpenNebulaModule):
    def __init__(self):
        argument_spec = dict(
            id=dict(type="int"),
            name=dict(type="str"),
            state=dict(type="str", choices=["present", "absent"], default="present"),
            template=dict(type="str"),
        )

        mutually_exclusive = [["id", "name"]]

        required_one_of = [("id", "name")]

        required_if = [["state", "present", ["template"]]]

        OpenNebulaModule.__init__(
            self,
            argument_spec,
            supports_check_mode=True,
            mutually_exclusive=mutually_exclusive,
            required_one_of=required_one_of,
            required_if=required_if,
        )

    def run(self, one, module, result):
        params = module.params
        id = params.get("id")
        name = params.get("name")
        desired_state = params.get("state")
        template_data = params.get("template")

        self.result = {}

        template = self.get_template_instance(id, name)
        needs_creation = False
        if not template and desired_state != "absent":
            if id:
                module.fail_json(msg=f"There is no template with id={id}")
            else:
                needs_creation = True

        if desired_state == "absent":
            self.result = self.delete_template(template)
        else:
            if needs_creation:
                self.result = self.create_template(name, template_data)
            else:
                self.result = self.update_template(template, template_data)

        self.exit()

    def get_template(self, predicate):
        # -2 means "Resources belonging to all users"
        # the other two parameters are used for pagination, -1 for both essentially means "return all"
        pool = self.one.vnpool.info(-2, -1, -1)

        for template in pool.VNET:
            if predicate(template):
                return template

        return None

    def get_template_by_id(self, template_id):
        return self.get_template(lambda template: template_id == template.ID)

    def get_template_by_name(self, name):
        return self.get_template(lambda template: name == template.NAME)

    def get_template_instance(self, requested_id, requested_name):
        if requested_id:
            return self.get_template_by_id(requested_id)
        else:
            return self.get_template_by_name(requested_name)

    def get_networks_ar_pool(self, template):
        ar_pool = []
        template_pool = template.AR_POOL.AR
        for ar in range(len(template_pool)):
            template_param = template_pool[ar]
            ar_pool.append(
                {
                    # These params will always be present
                    "ar_id": template_param.AR_ID,
                    "mac": template_param.MAC,
                    "size": template_param.SIZE,
                    "type": template_param.TYPE,
                    # These are optional so firstly check for presence
                    # and if not present set value to Null
                    "allocated": getattr(template_param, "ALLOCATED", "Null"),
                    "ip": getattr(template_param, "IP", "Null"),
                    "global_prefix": getattr(template_param, "GLOBAL_PREFIX", "Null"),
                    "parent_network_ar_id": getattr(template_param, "PARENT_NETWORK_AR_ID", "Null"),
                    "ula_prefix": getattr(template_param, "ULA_PREFIX", "Null"),
                    "vn_mad": getattr(template_param, "VN_MAD", "Null"),
                }
            )
        return ar_pool

    def get_template_info(self, template):
        info = {
            "id": template.ID,
            "name": template.NAME,
            "template": template.TEMPLATE,
            "user_name": template.UNAME,
            "user_id": template.UID,
            "group_name": template.GNAME,
            "group_id": template.GID,
            "permissions": {
                "owner_u": template.PERMISSIONS.OWNER_U,
                "owner_m": template.PERMISSIONS.OWNER_M,
                "owner_a": template.PERMISSIONS.OWNER_A,
                "group_u": template.PERMISSIONS.GROUP_U,
                "group_m": template.PERMISSIONS.GROUP_M,
                "group_a": template.PERMISSIONS.GROUP_A,
                "other_u": template.PERMISSIONS.OTHER_U,
                "other_m": template.PERMISSIONS.OTHER_M,
                "other_a": template.PERMISSIONS.OTHER_A,
            },
            "clusters": template.CLUSTERS.ID,
            "bridge": template.BRIDGE,
            "bride_type": template.BRIDGE_TYPE,
            "parent_network_id": template.PARENT_NETWORK_ID,
            "vn_mad": template.VN_MAD,
            "phydev": template.PHYDEV,
            "vlan_id": template.VLAN_ID,
            "outer_vlan_id": template.OUTER_VLAN_ID,
            "used_leases": template.USED_LEASES,
            "vrouters": template.VROUTERS.ID,
            "ar_pool": self.get_networks_ar_pool(template),
        }

        return info

    def create_template(self, name, template_data):
        if not self.module.check_mode:
            # -1 means that network won't be added to any cluster which happens by default
            self.one.vn.allocate(f'NAME = "{name}"\n{template_data}', -1)

        result = self.get_template_info(self.get_template_by_name(name))
        result["changed"] = True

        return result

    def update_template(self, template, template_data):
        if not self.module.check_mode:
            # 0 = replace the whole template
            self.one.vn.update(template.ID, template_data, 0)

        result = self.get_template_info(self.get_template_by_id(template.ID))
        if self.module.check_mode:
            # Unfortunately it is not easy to detect if the template would have changed, therefore always report a change here.
            result["changed"] = True
        else:
            # if the previous parsed template data is not equal to the updated one, this has changed
            result["changed"] = result["template"] != template.TEMPLATE

        return result

    def delete_template(self, template):
        if not template:
            return {"changed": False}

        if not self.module.check_mode:
            self.one.vn.delete(template.ID)

        return {"changed": True}


def main():
    NetworksModule().run_module()


if __name__ == "__main__":
    main()



#!/usr/bin/python
# Copyright (c) 2016, Werner Dijkerman (ikben@werner-dijkerman.nl)
# Copyright (c) 2017, Ansible Project
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

DOCUMENTATION = r"""
module: opendj_backendprop
short_description: Update the backend configuration of OpenDJ using the dsconfig set-backend-prop command
description:
  - This module updates settings for OpenDJ with the command C(set-backend-prop).
  - It checks first using C(get-backend-prop) if configuration needs to be applied.
author:
  - Werner Dijkerman (@dj-wasabi)
extends_documentation_fragment:
  - community.general._attributes
attributes:
  check_mode:
    support: full
  diff_mode:
    support: none
options:
  opendj_bindir:
    description:
      - The path to the bin directory of OpenDJ.
    default: /opt/opendj/bin
    type: path
  hostname:
    description:
      - The hostname of the OpenDJ server.
    required: true
    type: str
  port:
    description:
      - The Admin port on which the OpenDJ instance is available.
    required: true
    type: str
  username:
    description:
      - The username to connect with.
    default: cn=Directory Manager
    type: str
  password:
    description:
      - The password for O(username).
      - Either O(password) or O(passwordfile) is needed.
    type: str
  passwordfile:
    description:
      - Location to the password file which holds the password for O(username).
      - Either O(password) or O(passwordfile) is needed.
    type: path
  backend:
    description:
      - The name of the backend on which the property needs to be updated.
    required: true
    type: str
  name:
    description:
      - The configuration setting to update.
    required: true
    type: str
  value:
    description:
      - The value for the configuration item.
    required: true
    type: str
  state:
    description:
      - If configuration needs to be added/updated.
    default: "present"
    type: str
"""

EXAMPLES = r"""
- name: Add or update OpenDJ backend properties
  opendj_backendprop:
    hostname: localhost
    port: 4444
    username: "cn=Directory Manager"
    password: password
    backend: userRoot
    name: index-entry-limit
    value: 5000
"""

RETURN = r"""
"""

from ansible.module_utils.basic import AnsibleModule

from ansible_collections.community.general.plugins.module_utils._cmd_runner import CmdRunner, cmd_runner_fmt


class BackendProp:
    def __init__(self, module):
        self._module = module
        self.runner = CmdRunner(
            module,
            command=f"{module.params['opendj_bindir']}/dsconfig",
            arg_formats=dict(
                get_cmd=cmd_runner_fmt.as_fixed(["get-backend-prop", "-n", "-X", "-s"]),
                set_cmd=cmd_runner_fmt.as_fixed(["set-backend-prop", "-n", "-X"]),
                hostname=cmd_runner_fmt.as_opt_val("-h"),
                port=cmd_runner_fmt.as_opt_val("--port"),
                username=cmd_runner_fmt.as_opt_val("--bindDN"),
                password=cmd_runner_fmt.as_opt_val("-w"),
                passwordfile=cmd_runner_fmt.as_opt_val("-j"),
                backend=cmd_runner_fmt.as_opt_val("--backend-name"),
                set_prop=cmd_runner_fmt.as_func(cmd_runner_fmt.unpack_args(lambda n, v: ["--set", f"{n}:{v}"])),
            ),
        )

    def get_property(self):
        with self.runner(
            "get_cmd hostname port username password passwordfile backend",
            check_rc=True,
        ) as ctx:
            rc, stdout, stderr = ctx.run()
        return stdout

    def set_property(self, name, value):
        with self.runner(
            "set_cmd hostname port username password passwordfile backend set_prop",
            check_rc=True,
        ) as ctx:
            ctx.run(set_prop=[name, value])

    def validate_data(self, data, name, value):
        for config_line in data.splitlines():
            if config_line:
                split_line = config_line.split()
                if split_line[0] == name:
                    if split_line[1] == value:
                        return True
        return False


def main():
    module = AnsibleModule(
        argument_spec=dict(
            opendj_bindir=dict(default="/opt/opendj/bin", type="path"),
            hostname=dict(required=True),
            port=dict(required=True),
            username=dict(default="cn=Directory Manager"),
            password=dict(no_log=True),
            passwordfile=dict(type="path"),
            backend=dict(required=True),
            name=dict(required=True),
            value=dict(required=True),
            state=dict(default="present"),
        ),
        supports_check_mode=True,
        mutually_exclusive=[["password", "passwordfile"]],
        required_one_of=[["password", "passwordfile"]],
    )

    opendj = BackendProp(module)
    stdout = opendj.get_property()

    if stdout and not opendj.validate_data(data=stdout, name=module.params["name"], value=module.params["value"]):
        if module.check_mode:
            module.exit_json(changed=True)
        opendj.set_property(module.params["name"], module.params["value"])
        module.exit_json(changed=True)
    else:
        module.exit_json(changed=False)


if __name__ == "__main__":
    main()



#!/usr/bin/python

# Copyright (c) 2013, Patrik Lundin <patrik@sigterm.se>
#
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

DOCUMENTATION = r"""
module: openbsd_pkg
author:
  - Patrik Lundin (@eest)
short_description: Manage packages on OpenBSD
description:
  - Manage packages on OpenBSD using the pkg tools.
extends_documentation_fragment:
  - community.general._attributes
attributes:
  check_mode:
    support: full
  diff_mode:
    support: partial
    version_added: 9.1.0
    details:
      - Only works when check mode is not enabled.
options:
  name:
    description:
      - A name or a list of names of the packages.
    required: true
    type: list
    elements: str
  state:
    description:
      - V(present) ensures the package is installed.
      - V(latest) ensures the latest version of the package is installed.
      - V(absent) ensures the specified package is not installed.
    choices: [absent, latest, present, installed, removed]
    default: present
    type: str
  build:
    description:
      - Build the package from source instead of downloading and installing a binary. Requires that the port source tree is
        already installed. Automatically builds and installs the C(sqlports) package, if it is not already installed.
      - Mutually exclusive with O(snapshot).
    type: bool
    default: false
  snapshot:
    description:
      - Force C(%c) and C(%m) to expand to C(snapshots), even on a release kernel.
      - Mutually exclusive with O(build).
    type: bool
    default: false
    version_added: 1.3.0
  ports_dir:
    description:
      - When used in combination with the O(build) option, allows overriding the default ports source directory.
    default: /usr/ports
    type: path
  clean:
    description:
      - When updating or removing packages, delete the extra configuration file(s) in the old packages which are annotated
        with C(@extra) in the packaging-list.
    type: bool
    default: false
  quick:
    description:
      - Replace or delete packages quickly; do not bother with checksums before removing normal files.
    type: bool
    default: false
  autoremove:
    description:
      - Calls C(pkg_delete -a) to remove automatically installed packages which are no longer needed.
    type: bool
    default: false
    version_added: 11.3.0
notes:
  - When used with a C(loop:) each package is processed individually, it is much more efficient to pass the list directly
    to the O(name) option.
"""

EXAMPLES = r"""
- name: Make sure nmap is installed
  community.general.openbsd_pkg:
    name: nmap
    state: present

- name: Make sure nmap is the latest version
  community.general.openbsd_pkg:
    name: nmap
    state: latest

- name: Make sure nmap is not installed
  community.general.openbsd_pkg:
    name: nmap
    state: absent

- name: Make sure nmap is installed, build it from source if it is not
  community.general.openbsd_pkg:
    name: nmap
    state: present
    build: true

- name: Specify a pkg flavour with '--'
  community.general.openbsd_pkg:
    name: vim--no_x11
    state: present

- name: Specify the default flavour to avoid ambiguity errors
  community.general.openbsd_pkg:
    name: vim--
    state: present

- name: Specify a package branch (requires at least OpenBSD 6.0)
  community.general.openbsd_pkg:
    name: python%3.5
    state: present

- name: Update all packages on the system
  community.general.openbsd_pkg:
    name: '*'
    state: latest

- name: Purge a package and its configuration files
  community.general.openbsd_pkg:
    name: mpd
    clean: true
    state: absent

- name: Quickly remove a package without checking checksums
  community.general.openbsd_pkg:
    name: qt5
    quick: true
    state: absent

- name: Install packages, remove unused dependencies
  community.general.openbsd_pkg:
    name: ["tree", "mtr"]
    autoremove: true

- name: Remove all unused dependencies
  community.general.openbsd_pkg:
    name: '*'
    autoremove: true
"""

import os
import platform
import re
import shlex
import sqlite3

from ansible.module_utils.basic import AnsibleModule

from ansible_collections.community.general.plugins.module_utils._version import LooseVersion


# Function used for executing commands.
def execute_command(cmd, module):
    # Break command line into arguments.
    # This makes run_command() use shell=False which we need to not cause shell
    # expansion of special characters like '*'.
    cmd_args = shlex.split(cmd)

    # We set TERM to 'dumb' to keep pkg_add happy if the machine running
    # ansible is using a TERM that the managed machine does not know about,
    # e.g.: "No progress meter: failed termcap lookup on xterm-kitty".
    return module.run_command(cmd_args, environ_update={"TERM": "dumb", "LANGUAGE": "C", "LC_ALL": "C"})


def get_all_installed(module):
    """
    Get all installed packaged. Used to support diff mode
    """
    command = "pkg_info -Iq"

    rc, stdout, stderr = execute_command(command, module)

    if stderr:
        module.fail_json(msg=f"failed in get_all_installed(): {stderr}")

    return stdout


# Function used to find out if a package is currently installed.
def get_package_state(names, pkg_spec, module):
    info_cmd = "pkg_info -Iq"

    for name in names:
        command = f"{info_cmd} inst:{name}"

        rc, stdout, stderr = execute_command(command, module)

        if stderr:
            match = re.search(rf"^Can't find inst:{re.escape(name)}$", stderr)
            if match:
                pkg_spec[name]["installed_state"] = False
            else:
                module.fail_json(msg=f"failed in get_package_state(): {stderr}")

        if stdout:
            # If the requested package name is just a stem, like "python", we may
            # find multiple packages with that name.
            pkg_spec[name]["installed_names"] = stdout.splitlines()
            module.debug(f"get_package_state(): installed_names = {pkg_spec[name]['installed_names']}")
            pkg_spec[name]["installed_state"] = True
        else:
            pkg_spec[name]["installed_state"] = False


# Function used to make sure a package is present.
def package_present(names, pkg_spec, module):
    build = module.params["build"]

    for name in names:
        # It is possible package_present() has been called from package_latest().
        # In that case we do not want to operate on the whole list of names,
        # only the leftovers.
        if pkg_spec["package_latest_leftovers"]:
            if name not in pkg_spec["package_latest_leftovers"]:
                module.debug(f"package_present(): ignoring '{name}' which is not a package_latest() leftover")
                continue
            else:
                module.debug(f"package_present(): handling package_latest() leftovers, installing '{name}'")

        if module.check_mode:
            install_cmd = "pkg_add -Imn"
        else:
            if build is True:
                port_dir = f"{module.params['ports_dir']}/{get_package_source_path(name, pkg_spec, module)}"
                if os.path.isdir(port_dir):
                    if pkg_spec[name]["flavor"]:
                        flavors = pkg_spec[name]["flavor"].replace("-", " ")
                        install_cmd = f'cd {port_dir} && make clean=depends && FLAVOR="{flavors}" make install && make clean=depends'
                    elif pkg_spec[name]["subpackage"]:
                        install_cmd = f'cd {port_dir} && make clean=depends && SUBPACKAGE="{pkg_spec[name]["subpackage"]}" make install && make clean=depends'
                    else:
                        install_cmd = f"cd {port_dir} && make install && make clean=depends"
                else:
                    module.fail_json(msg=f"the port source directory {port_dir} does not exist")
            else:
                install_cmd = "pkg_add -Im"

        if module.params["snapshot"] is True:
            install_cmd += " -Dsnap"

        if pkg_spec[name]["installed_state"] is False:
            # Attempt to install the package
            if build is True and not module.check_mode:
                (pkg_spec[name]["rc"], pkg_spec[name]["stdout"], pkg_spec[name]["stderr"]) = module.run_command(
                    install_cmd, module, use_unsafe_shell=True
                )
            else:
                (pkg_spec[name]["rc"], pkg_spec[name]["stdout"], pkg_spec[name]["stderr"]) = execute_command(
                    f"{install_cmd} {name}", module
                )

            # The behaviour of pkg_add is a bit different depending on if a
            # specific version is supplied or not.
            #
            # When a specific version is supplied the return code will be 0 when
            # a package is found and 1 when it is not. If a version is not
            # supplied the tool will exit 0 in both cases.
            #
            # It is important to note that "version" relates to the
            # packages-specs(7) notion of a version. If using the branch syntax
            # (like "python%3.5") even though a branch name may look like a
            # version string it is not used an one by pkg_add.
            if pkg_spec[name]["version"] or build is True:
                # Depend on the return code.
                module.debug(f"package_present(): depending on return code for name '{name}'")
                if pkg_spec[name]["rc"]:
                    pkg_spec[name]["changed"] = False
            else:
                # Depend on stderr instead.
                module.debug(f"package_present(): depending on stderr for name '{name}'")
                if pkg_spec[name]["stderr"]:
                    # There is a corner case where having an empty directory in
                    # installpath prior to the right location will result in a
                    # "file:/local/package/directory/ is empty" message on stderr
                    # while still installing the package, so we need to look for
                    # for a message like "packagename-1.0: ok" just in case.
                    match = re.search(rf"\W{re.escape(pkg_spec[name]['stem'])}-[^:]+: ok\W", pkg_spec[name]["stdout"])

                    if match:
                        # It turns out we were able to install the package.
                        module.debug(f"package_present(): we were able to install package for name '{name}'")
                        pkg_spec[name]["changed"] = True
                    else:
                        # We really did fail, fake the return code.
                        module.debug(f"package_present(): we really did fail for name '{name}'")
                        pkg_spec[name]["rc"] = 1
                        pkg_spec[name]["changed"] = False
                else:
                    module.debug(f"package_present(): stderr was not set for name '{name}'")

            if pkg_spec[name]["rc"] == 0:
                pkg_spec[name]["changed"] = True

        else:
            pkg_spec[name]["rc"] = 0
            pkg_spec[name]["stdout"] = ""
            pkg_spec[name]["stderr"] = ""
            pkg_spec[name]["changed"] = False


# Function used to make sure a package is the latest available version.
def package_latest(names, pkg_spec, module):
    if module.params["build"] is True:
        module.fail_json(msg=f"the combination of build={module.params['build']} and state=latest is not supported")

    upgrade_cmd = "pkg_add -um"

    if module.check_mode:
        upgrade_cmd += "n"

    if module.params["clean"]:
        upgrade_cmd += "c"

    if module.params["quick"]:
        upgrade_cmd += "q"

    if module.params["snapshot"]:
        upgrade_cmd += " -Dsnap"

    for name in names:
        if pkg_spec[name]["installed_state"] is True:
            # Attempt to upgrade the package.
            (pkg_spec[name]["rc"], pkg_spec[name]["stdout"], pkg_spec[name]["stderr"]) = execute_command(
                f"{upgrade_cmd} {name}", module
            )

            # Look for output looking something like "nmap-6.01->6.25: ok" to see if
            # something changed (or would have changed). Use \W to delimit the match
            # from progress meter output.
            pkg_spec[name]["changed"] = False
            for installed_name in pkg_spec[name]["installed_names"]:
                module.debug(f"package_latest(): checking for pre-upgrade package name: {installed_name}")
                match = re.search(rf"\W{re.escape(installed_name)}->.+: ok\W", pkg_spec[name]["stdout"])
                if match:
                    module.debug(f"package_latest(): pre-upgrade package name match: {installed_name}")

                    pkg_spec[name]["changed"] = True
                    break

            # FIXME: This part is problematic. Based on the issues mentioned (and
            # handled) in package_present() it is not safe to blindly trust stderr
            # as an indicator that the command failed, and in the case with
            # empty installpath directories this will break.
            #
            # For now keep this safeguard here, but ignore it if we managed to
            # parse out a successful update above. This way we will report a
            # successful run when we actually modify something but fail
            # otherwise.
            if pkg_spec[name]["changed"] is not True:
                if pkg_spec[name]["stderr"]:
                    pkg_spec[name]["rc"] = 1

        else:
            # Note packages that need to be handled by package_present
            module.debug(f"package_latest(): package '{name}' is not installed, will be handled by package_present()")
            pkg_spec["package_latest_leftovers"].append(name)

    # If there were any packages that were not installed we call
    # package_present() which will handle those.
    if pkg_spec["package_latest_leftovers"]:
        module.debug("package_latest(): calling package_present() to handle leftovers")
        package_present(names, pkg_spec, module)


# Function used to make sure a package is not installed.
def package_absent(names, pkg_spec, module):
    remove_cmd = "pkg_delete -I"

    if module.check_mode:
        remove_cmd += "n"

    if module.params["clean"]:
        remove_cmd += "c"

    if module.params["quick"]:
        remove_cmd += "q"

    for name in names:
        if pkg_spec[name]["installed_state"] is True:
            # Attempt to remove the package.
            (pkg_spec[name]["rc"], pkg_spec[name]["stdout"], pkg_spec[name]["stderr"]) = execute_command(
                f"{remove_cmd} {name}", module
            )

            if pkg_spec[name]["rc"] == 0:
                pkg_spec[name]["changed"] = True
            else:
                pkg_spec[name]["changed"] = False

        else:
            pkg_spec[name]["rc"] = 0
            pkg_spec[name]["stdout"] = ""
            pkg_spec[name]["stderr"] = ""
            pkg_spec[name]["changed"] = False


# Function used to remove unused dependencies.
def package_rm_unused_deps(pkg_spec, module):
    rm_unused_deps_cmd = "pkg_delete -Ia"

    if module.check_mode:
        rm_unused_deps_cmd += "n"

    if module.params["clean"]:
        rm_unused_deps_cmd += "c"

    if module.params["quick"]:
        rm_unused_deps_cmd += "q"

    # If we run the commands, we set changed to true to let
    # the package list change detection code do the actual work.

    # Create a minimal pkg_spec entry for '*' to store return values.
    pkg_spec["*"] = {}

    # Attempt to remove unused dependencies.
    pkg_spec["*"]["rc"], pkg_spec["*"]["stdout"], pkg_spec["*"]["stderr"] = execute_command(rm_unused_deps_cmd, module)
    pkg_spec["*"]["changed"] = True


# Function used to parse the package name based on packages-specs(7).
# The general name structure is "stem-version[-flavors]".
#
# Names containing "%" are a special variation not part of the
# packages-specs(7) syntax. See pkg_add(1) on OpenBSD 6.0 or later for a
# description.
def parse_package_name(names, pkg_spec, module):
    # Initialize empty list of package_latest() leftovers.
    pkg_spec["package_latest_leftovers"] = []

    for name in names:
        module.debug(f"parse_package_name(): parsing name: {name}")
        # Do some initial matches so we can base the more advanced regex on that.
        version_match = re.search("-[0-9]", name)
        versionless_match = re.search("--", name)

        # Stop if someone is giving us a name that both has a version and is
        # version-less at the same time.
        if version_match and versionless_match:
            module.fail_json(msg=f"package name both has a version and is version-less: {name}")

        # All information for a given name is kept in the pkg_spec keyed by that name.
        pkg_spec[name] = {}

        # If name includes a version.
        if version_match:
            match = re.search(
                "^(?P<stem>[^%]+)-(?P<version>[0-9][^-]*)(?P<flavor_separator>-)?(?P<flavor>[a-z].*)?(%(?P<branch>.+))?$",
                name,
            )
            if match:
                pkg_spec[name]["stem"] = match.group("stem")
                pkg_spec[name]["version_separator"] = "-"
                pkg_spec[name]["version"] = match.group("version")
                pkg_spec[name]["flavor_separator"] = match.group("flavor_separator")
                pkg_spec[name]["flavor"] = match.group("flavor")
                pkg_spec[name]["branch"] = match.group("branch")
                pkg_spec[name]["style"] = "version"
                module.debug(
                    f"version_match: stem: {pkg_spec[name]['stem']}, version: {pkg_spec[name]['version']}, "
                    f"flavor_separator: {pkg_spec[name]['flavor_separator']}, flavor: {pkg_spec[name]['flavor']}, branch: {pkg_spec[name]['branch']}, "
                    f"style: {pkg_spec[name]['style']}"
                )
            else:
                module.fail_json(msg=f"unable to parse package name at version_match: {name}")

        # If name includes no version but is version-less ("--").
        elif versionless_match:
            match = re.search("^(?P<stem>[^%]+)--(?P<flavor>[a-z].*)?(%(?P<branch>.+))?$", name)
            if match:
                pkg_spec[name]["stem"] = match.group("stem")
                pkg_spec[name]["version_separator"] = "-"
                pkg_spec[name]["version"] = None
                pkg_spec[name]["flavor_separator"] = "-"
                pkg_spec[name]["flavor"] = match.group("flavor")
                pkg_spec[name]["branch"] = match.group("branch")
                pkg_spec[name]["style"] = "versionless"
                module.debug(
                    f"versionless_match: stem: {pkg_spec[name]['stem']}, flavor: {pkg_spec[name]['flavor']}, "
                    f"branch: {pkg_spec[name]['branch']}, style: {pkg_spec[name]['style']}"
                )
            else:
                module.fail_json(msg=f"unable to parse package name at versionless_match: {name}")

        # If name includes no version, and is not version-less, it is all a
        # stem, possibly with a branch (%branchname) tacked on at the
        # end.
        else:
            match = re.search("^(?P<stem>[^%]+)(%(?P<branch>.+))?$", name)
            if match:
                pkg_spec[name]["stem"] = match.group("stem")
                pkg_spec[name]["version_separator"] = None
                pkg_spec[name]["version"] = None
                pkg_spec[name]["flavor_separator"] = None
                pkg_spec[name]["flavor"] = None
                pkg_spec[name]["branch"] = match.group("branch")
                pkg_spec[name]["style"] = "stem"
                module.debug(
                    f"stem_match: stem: {pkg_spec[name]['stem']}, branch: {pkg_spec[name]['branch']}, style: {pkg_spec[name]['style']}"
                )
            else:
                module.fail_json(msg=f"unable to parse package name at else: {name}")

        # Verify that the managed host is new enough to support branch syntax.
        if pkg_spec[name]["branch"]:
            branch_release = "6.0"

            if LooseVersion(platform.release()) < LooseVersion(branch_release):
                module.fail_json(
                    msg=f"package name using 'branch' syntax requires at least OpenBSD {branch_release}: {name}"
                )

        # Sanity check that there are no trailing dashes in flavor.
        # Try to stop strange stuff early so we can be strict later.
        if pkg_spec[name]["flavor"]:
            match = re.search("-$", pkg_spec[name]["flavor"])
            if match:
                module.fail_json(msg=f"trailing dash in flavor: {pkg_spec[name]['flavor']}")


# Function used for figuring out the port path.
def get_package_source_path(name, pkg_spec, module):
    pkg_spec[name]["subpackage"] = None
    if pkg_spec[name]["stem"] == "sqlports":
        return "databases/sqlports"
    else:
        # try for an exact match first
        sqlports_db_file = "/usr/local/share/sqlports"
        if not os.path.isfile(sqlports_db_file):
            module.fail_json(msg=f"sqlports file '{sqlports_db_file}' is missing")

        conn = sqlite3.connect(sqlports_db_file)
        first_part_of_query = "SELECT fullpkgpath, fullpkgname FROM ports WHERE fullpkgname"
        query = f"{first_part_of_query} = ?"
        module.debug(f"package_package_source_path(): exact query: {query}")
        cursor = conn.execute(query, (name,))
        results = cursor.fetchall()

        # next, try for a fuzzier match
        if len(results) < 1:
            looking_for = (
                pkg_spec[name]["stem"]
                + (pkg_spec[name]["version_separator"] or "-")
                + (pkg_spec[name]["version"] or "%")
            )
            query = f"{first_part_of_query} LIKE ?"
            if pkg_spec[name]["flavor"]:
                looking_for += pkg_spec[name]["flavor_separator"] + pkg_spec[name]["flavor"]
                module.debug(f"package_package_source_path(): fuzzy flavor query: {query}")
                cursor = conn.execute(query, (looking_for,))
            elif pkg_spec[name]["style"] == "versionless":
                query += " AND fullpkgname NOT LIKE ?"
                module.debug(f"package_package_source_path(): fuzzy versionless query: {query}")
                cursor = conn.execute(
                    query,
                    (
                        looking_for,
                        f"{looking_for}-%",
                    ),
                )
            else:
                module.debug(f"package_package_source_path(): fuzzy query: {query}")
                cursor = conn.execute(query, (looking_for,))
            results = cursor.fetchall()

        # error if we don't find exactly 1 match
        conn.close()
        if len(results) < 1:
            module.fail_json(msg=f"could not find a port by the name '{name}'")
        if len(results) > 1:
            matches = map(lambda x: x[1], results)
            module.fail_json(msg=f"too many matches, unsure which to build: {' OR '.join(matches)}")

        # there's exactly 1 match, so figure out the subpackage, if any, then return
        fullpkgpath = results[0][0]
        parts = fullpkgpath.split(",")
        if len(parts) > 1 and parts[1][0] == "-":
            pkg_spec[name]["subpackage"] = parts[1]
        return parts[0]


# Function used for upgrading all installed packages.
def upgrade_packages(pkg_spec, module):
    if module.check_mode:
        upgrade_cmd = "pkg_add -Imnu"
    else:
        upgrade_cmd = "pkg_add -Imu"

    if module.params["snapshot"]:
        upgrade_cmd += " -Dsnap"

    # Create a minimal pkg_spec entry for '*' to store return values.
    pkg_spec["*"] = {}

    # Attempt to upgrade all packages.
    pkg_spec["*"]["rc"], pkg_spec["*"]["stdout"], pkg_spec["*"]["stderr"] = execute_command(upgrade_cmd, module)

    # Try to find any occurrence of a package changing version like:
    # "bzip2-1.0.6->1.0.6p0: ok".
    match = re.search(r"\W\w.+->.+: ok\W", pkg_spec["*"]["stdout"])
    if match:
        pkg_spec["*"]["changed"] = True

    else:
        pkg_spec["*"]["changed"] = False

    # It seems we can not trust the return value, so depend on the presence of
    # stderr to know if something failed.
    if pkg_spec["*"]["stderr"]:
        pkg_spec["*"]["rc"] = 1
    else:
        pkg_spec["*"]["rc"] = 0


# ===========================================
# Main control flow.
def main():
    module = AnsibleModule(
        argument_spec=dict(
            name=dict(type="list", elements="str", required=True),
            state=dict(type="str", default="present", choices=["absent", "installed", "latest", "present", "removed"]),
            build=dict(type="bool", default=False),
            snapshot=dict(type="bool", default=False),
            ports_dir=dict(type="path", default="/usr/ports"),
            quick=dict(type="bool", default=False),
            clean=dict(type="bool", default=False),
            autoremove=dict(type="bool", default=False),
        ),
        mutually_exclusive=[["snapshot", "build"]],
        supports_check_mode=True,
    )

    name = module.params["name"]
    state = module.params["state"]
    build = module.params["build"]
    ports_dir = module.params["ports_dir"]

    result = {}
    result["name"] = name
    result["state"] = state
    result["build"] = build
    result["diff"] = {}

    # The data structure used to keep track of package information.
    pkg_spec = {}

    new_package_list = original_package_list = get_all_installed(module)

    if build is True:
        if not os.path.isdir(ports_dir):
            module.fail_json(msg=f"the ports source directory {ports_dir} does not exist")

        # build sqlports if its not installed yet
        parse_package_name(["sqlports"], pkg_spec, module)
        get_package_state(["sqlports"], pkg_spec, module)
        if not pkg_spec["sqlports"]["installed_state"]:
            module.debug(f"main(): installing 'sqlports' because build={module.params['build']}")
            package_present(["sqlports"], pkg_spec, module)

    asterisk_name = False
    for n in name:
        if n == "*":
            if len(name) != 1:
                module.fail_json(msg="the package name '*' can not be mixed with other names")

            asterisk_name = True

    if asterisk_name:
        if state != "latest" and not module.params["autoremove"]:
            module.fail_json(msg="the package name '*' is only valid when using state=latest or autoremove=true")

        if state == "latest":
            # Perform an upgrade of all installed packages.
            upgrade_packages(pkg_spec, module)

        if module.params["autoremove"]:
            # Remove unused dependencies.
            package_rm_unused_deps(pkg_spec, module)
    else:
        # Parse package names and put results in the pkg_spec dictionary.
        parse_package_name(name, pkg_spec, module)

        # Not sure how the branch syntax is supposed to play together
        # with build mode. Disable it for now.
        for n in name:
            if pkg_spec[n]["branch"] and module.params["build"] is True:
                module.fail_json(
                    msg=f"the combination of 'branch' syntax and build={module.params['build']} is not supported: {n}"
                )

        # Get state for all package names.
        get_package_state(name, pkg_spec, module)

        # Perform requested action.
        if state in ["installed", "present"]:
            package_present(name, pkg_spec, module)
        elif state in ["absent", "removed"]:
            package_absent(name, pkg_spec, module)
        elif state == "latest":
            package_latest(name, pkg_spec, module)

        # Handle autoremove if requested for non-asterisk packages
        if module.params["autoremove"]:
            package_rm_unused_deps(pkg_spec, module)

    # The combined changed status for all requested packages. If anything
    # is changed this is set to True.
    combined_changed = False

    # The combined failed status for all requested packages. If anything
    # failed this is set to True.
    combined_failed = False

    # We combine all error messages in this comma separated string, for example:
    # "msg": "Can't find nmapp\n, Can't find nmappp\n"
    combined_error_message = ""

    # Loop over all requested package names and check if anything failed or
    # changed.
    for n in name:
        if pkg_spec[n]["rc"] != 0:
            combined_failed = True
            if pkg_spec[n]["stderr"]:
                if combined_error_message:
                    combined_error_message += f", {pkg_spec[n]['stderr']}"
                else:
                    combined_error_message = pkg_spec[n]["stderr"]
            else:
                if combined_error_message:
                    combined_error_message += f", {pkg_spec[n]['stdout']}"
                else:
                    combined_error_message = pkg_spec[n]["stdout"]

        if pkg_spec[n]["changed"] is True:
            combined_changed = True

    # If combined_error_message contains anything at least some part of the
    # list of requested package names failed.
    if combined_failed:
        module.fail_json(msg=combined_error_message, **result)

    result["changed"] = combined_changed

    if not module.check_mode:
        new_package_list = get_all_installed(module)
        result["diff"] = dict(before=original_package_list, after=new_package_list)
        result["changed"] = result["diff"]["before"] != result["diff"]["after"]

    module.exit_json(**result)


if __name__ == "__main__":
    main()



#!/usr/bin/python

# Copyright Ansible Project
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

DOCUMENTATION = r"""
module: vertica_role
short_description: Adds or removes Vertica database roles and assigns roles to them
description:
  - Adds or removes Vertica database role and, optionally, assign other roles.
extends_documentation_fragment:
  - community.general._attributes
attributes:
  check_mode:
    support: full
  diff_mode:
    support: none
options:
  role:
    description:
      - Name of the role to add or remove.
    required: true
    type: str
    aliases: ['name']
  assigned_roles:
    description:
      - Comma separated list of roles to assign to the role.
    aliases: ['assigned_role']
    type: str
  state:
    description:
      - Whether to create V(present), drop V(absent) or lock V(locked) a role.
    choices: ['present', 'absent']
    default: present
    type: str
  db:
    description:
      - Name of the Vertica database.
    type: str
  cluster:
    description:
      - Name of the Vertica cluster.
    default: localhost
    type: str
  port:
    description:
      - Vertica cluster port to connect to.
    default: '5433'
    type: str
  login_user:
    description:
      - The username used to authenticate with.
    default: dbadmin
    type: str
  login_password:
    description:
      - The password used to authenticate with.
    type: str
notes:
  - The default authentication assumes that you are either logging in as or sudo'ing to the C(dbadmin) account on the host.
  - This module uses C(pyodbc), a Python ODBC database adapter. You must ensure that C(unixODBC) and C(pyodbc) is installed
    on the host and properly configured.
  - Configuring C(unixODBC) for Vertica requires C(Driver = /opt/vertica/lib64/libverticaodbc.so) to be added to the C(Vertica)
    section of either C(/etc/odbcinst.ini) or C($HOME/.odbcinst.ini) and both C(ErrorMessagesPath = /opt/vertica/lib64) and
    C(DriverManagerEncoding = UTF-16) to be added to the C(Driver) section of either C(/etc/vertica.ini) or C($HOME/.vertica.ini).
requirements: ['unixODBC', 'pyodbc']
author: "Dariusz Owczarek (@dareko)"
"""

EXAMPLES = r"""
- name: Creating a new vertica role
  community.general.vertica_role: name=role_name db=db_name state=present

- name: Creating a new vertica role with other role assigned
  community.general.vertica_role: name=role_name assigned_role=other_role_name state=present
"""
import traceback

PYODBC_IMP_ERR = None
try:
    import pyodbc
except ImportError:
    PYODBC_IMP_ERR = traceback.format_exc()
    pyodbc_found = False
else:
    pyodbc_found = True

from ansible.module_utils.basic import AnsibleModule, missing_required_lib


class NotSupportedError(Exception):
    pass


class CannotDropError(Exception):
    pass


# module specific functions


def get_role_facts(cursor, role=""):
    facts = {}
    cursor.execute(
        """
        select r.name, r.assigned_roles
        from roles r
        where (? = '' or r.name ilike ?)
    """,
        role,
        role,
    )
    while True:
        rows = cursor.fetchmany(100)
        if not rows:
            break
        for row in rows:
            role_key = row.name.lower()
            facts[role_key] = {"name": row.name, "assigned_roles": []}
            if row.assigned_roles:
                facts[role_key]["assigned_roles"] = row.assigned_roles.replace(" ", "").split(",")
    return facts


def update_roles(role_facts, cursor, role, existing, required):
    for assigned_role in set(existing) - set(required):
        cursor.execute(f"revoke {assigned_role} from {role}")
    for assigned_role in set(required) - set(existing):
        cursor.execute(f"grant {assigned_role} to {role}")


def check(role_facts, role, assigned_roles):
    role_key = role.lower()
    if role_key not in role_facts:
        return False
    return not (assigned_roles and sorted(assigned_roles) != sorted(role_facts[role_key]["assigned_roles"]))


def present(role_facts, cursor, role, assigned_roles):
    role_key = role.lower()
    if role_key not in role_facts:
        cursor.execute(f"create role {role}")
        update_roles(role_facts, cursor, role, [], assigned_roles)
        role_facts.update(get_role_facts(cursor, role))
        return True
    else:
        changed = False
        if assigned_roles and (sorted(assigned_roles) != sorted(role_facts[role_key]["assigned_roles"])):
            update_roles(role_facts, cursor, role, role_facts[role_key]["assigned_roles"], assigned_roles)
            changed = True
        if changed:
            role_facts.update(get_role_facts(cursor, role))
        return changed


def absent(role_facts, cursor, role, assigned_roles):
    role_key = role.lower()
    if role_key in role_facts:
        update_roles(role_facts, cursor, role, role_facts[role_key]["assigned_roles"], [])
        cursor.execute(f"drop role {role_facts[role_key]['name']} cascade")
        del role_facts[role_key]
        return True
    else:
        return False


# module logic


def main():
    module = AnsibleModule(
        argument_spec=dict(
            role=dict(required=True, aliases=["name"]),
            assigned_roles=dict(aliases=["assigned_role"]),
            state=dict(default="present", choices=["absent", "present"]),
            db=dict(),
            cluster=dict(default="localhost"),
            port=dict(default="5433"),
            login_user=dict(default="dbadmin"),
            login_password=dict(no_log=True),
        ),
        supports_check_mode=True,
    )

    if not pyodbc_found:
        module.fail_json(msg=missing_required_lib("pyodbc"), exception=PYODBC_IMP_ERR)

    role = module.params["role"]
    assigned_roles = []
    if module.params["assigned_roles"]:
        assigned_roles = module.params["assigned_roles"].split(",")
        assigned_roles = [_f for _f in assigned_roles if _f]
    state = module.params["state"]
    db = ""
    if module.params["db"]:
        db = module.params["db"]

    changed = False

    try:
        dsn = (
            "Driver=Vertica;"
            f"Server={module.params['cluster']};"
            f"Port={module.params['port']};"
            f"Database={db};"
            f"User={module.params['login_user']};"
            f"Password={module.params['login_password']};"
            f"ConnectionLoadBalance=true"
        )
        db_conn = pyodbc.connect(dsn, autocommit=True)
        cursor = db_conn.cursor()
    except Exception as e:
        module.fail_json(msg=f"Unable to connect to database: {e}.")

    try:
        role_facts = get_role_facts(cursor)
        if module.check_mode:
            changed = not check(role_facts, role, assigned_roles)
        elif state == "absent":
            try:
                changed = absent(role_facts, cursor, role, assigned_roles)
            except pyodbc.Error as e:
                module.fail_json(msg=f"{e}", exception=traceback.format_exc())
        elif state == "present":
            try:
                changed = present(role_facts, cursor, role, assigned_roles)
            except pyodbc.Error as e:
                module.fail_json(msg=f"{e}", exception=traceback.format_exc())
    except NotSupportedError as e:
        module.fail_json(msg=f"{e}", ansible_facts={"vertica_roles": role_facts})
    except CannotDropError as e:
        module.fail_json(msg=f"{e}", ansible_facts={"vertica_roles": role_facts})
    except SystemExit:
        # avoid catching this on python 2.4
        raise
    except Exception as e:
        module.fail_json(msg=f"{e}", exception=traceback.format_exc())

    module.exit_json(changed=changed, role=role, ansible_facts={"vertica_roles": role_facts})


if __name__ == "__main__":
    main()



#!/usr/bin/python

# Copyright (c) 2018, Juergen Wiebe <wiebe@e-spirit.com>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later
from __future__ import annotations

DOCUMENTATION = r"""
module: utm_network_interface_address

author:
  - Juergen Wiebe (@steamx)

short_description: Create, update or destroy network/interface_address object

description:
  - Create, update or destroy a network/interface_address object in SOPHOS UTM.
  - This module needs to have the REST Ability of the UTM to be activated.
attributes:
  check_mode:
    support: none
  diff_mode:
    support: none

options:
  name:
    type: str
    description:
      - The name of the object that identifies the entry.
    required: true
  address:
    type: str
    description:
      - The ip4 address of the network/interface_address object.
    required: true
  address6:
    type: str
    description:
      - The ip6 address of the network/interface_address object.
  comment:
    type: str
    description:
      - An optional comment to add to the object.
    default: ''
  resolved:
    type: bool
    description:
      - Whether or not the object is resolved.
  resolved6:
    type: bool
    description:
      - Whether or not the object is resolved.
extends_documentation_fragment:
  - community.general._utm
  - community.general._attributes
"""

EXAMPLES = r"""
- name: Create a network interface address
  utm_proxy_backend:
    utm_host: sophos.host.name
    utm_token: abcdefghijklmno1234
    name: TestNetworkInterfaceAddress
    address: 0.0.0.0
    state: present

- name: Remove a network interface address
  network_interface_address:
    utm_host: sophos.host.name
    utm_token: abcdefghijklmno1234
    name: TestNetworkInterfaceAddress
    address: 0.0.0.0
    state: absent
"""

RETURN = r"""
result:
  description: The utm object that was created.
  returned: success
  type: complex
  contains:
    _ref:
      description: The reference name of the object.
      type: str
    _locked:
      description: Whether or not the object is currently locked.
      type: bool
    _type:
      description: The type of the object.
      type: str
    name:
      description: The name of the object.
      type: str
    address:
      description: The ip4 address of the network/interface_address object.
      type: str
    address6:
      description: The ip6 address of the network/interface_address object.
      type: str
    comment:
      description: The comment string.
      type: str
    resolved:
      description: Whether or not the object is resolved.
      type: bool
    resolved6:
      description: Whether or not the object is resolved.
      type: bool
"""

from ansible_collections.community.general.plugins.module_utils._utm_utils import UTM, UTMModule


def main():
    endpoint = "network/interface_address"
    key_to_check_for_changes = ["comment", "address"]
    module = UTMModule(
        argument_spec=dict(
            name=dict(type="str", required=True),
            address=dict(type="str", required=True),
            comment=dict(type="str", default=""),
            address6=dict(type="str"),
            resolved=dict(type="bool"),
            resolved6=dict(type="bool"),
        )
    )
    try:
        UTM(module, endpoint, key_to_check_for_changes).execute()
    except Exception as e:
        module.fail_json(msg=f"{e}")


if __name__ == "__main__":
    main()



#!/usr/bin/python
#
# Scaleway volumes management module
#
# Copyright (C) 2018 Henryk Konsek Consulting (hekonsek@gmail.com).
#
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

DOCUMENTATION = r"""
module: scaleway_volume
short_description: Scaleway volumes management module
author: Henryk Konsek (@hekonsek)
description:
  - This module manages volumes on Scaleway account U(https://developer.scaleway.com).
extends_documentation_fragment:
  - community.general._scaleway
  - community.general._attributes
  - community.general._scaleway.actiongroup_scaleway

attributes:
  check_mode:
    support: full
  diff_mode:
    support: none
  action_group:
    version_added: 11.3.0

options:
  state:
    type: str
    description:
      - Indicate desired state of the volume.
    default: present
    choices:
      - present
      - absent
  region:
    type: str
    description:
      - Scaleway region to use (for example par1).
    required: true
    choices:
      - ams1
      - EMEA-NL-EVS
      - ams2
      - ams3
      - par1
      - EMEA-FR-PAR1
      - par2
      - EMEA-FR-PAR2
      - par3
      - waw1
      - EMEA-PL-WAW1
      - waw2
      - waw3
  name:
    type: str
    description:
      - Name used to identify the volume.
    required: true
  project:
    type: str
    description:
      - Scaleway project ID to which volume belongs.
    version_added: 4.3.0
  organization:
    type: str
    description:
      - ScaleWay organization ID to which volume belongs.
  size:
    type: int
    description:
      - Size of the volume in bytes.
  volume_type:
    type: str
    description:
      - Type of the volume (for example 'l_ssd').
"""

EXAMPLES = r"""
- name: Create 10GB volume
  community.general.scaleway_volume:
    name: my-volume
    state: present
    region: par1
    project: "{{ scw_org }}"
    "size": 10000000000
    volume_type: l_ssd
  register: server_creation_check_task

- name: Make sure volume deleted
  community.general.scaleway_volume:
    name: my-volume
    state: absent
    region: par1
"""

RETURN = r"""
data:
  description: This is only present when O(state=present).
  returned: when O(state=present)
  type: dict
  sample:
    {
      "volume": {
        "export_uri": null,
        "id": "c675f420-cfeb-48ff-ba2a-9d2a4dbe3fcd",
        "name": "volume-0-3",
        "project": "000a115d-2852-4b0a-9ce8-47f1134ba95a",
        "server": null,
        "size": 10000000000,
        "volume_type": "l_ssd"
      }
    }
"""

from ansible.module_utils.basic import AnsibleModule

from ansible_collections.community.general.plugins.module_utils._scaleway import (
    SCALEWAY_LOCATION,
    Scaleway,
    scaleway_argument_spec,
)


def core(module):
    region = module.params["region"]
    state = module.params["state"]
    name = module.params["name"]
    organization = module.params["organization"]
    project = module.params["project"]
    size = module.params["size"]
    volume_type = module.params["volume_type"]
    module.params["api_url"] = SCALEWAY_LOCATION[region]["api_endpoint"]

    account_api = Scaleway(module)
    response = account_api.get("volumes")
    status_code = response.status_code
    volumes_json = response.json

    if project is None:
        project = organization

    if not response.ok:
        module.fail_json(msg=f"Error getting volume [{status_code}: {response.json['message']}]")

    volumeByName = None
    for volume in volumes_json["volumes"]:
        if volume["project"] == project and volume["name"] == name:
            volumeByName = volume

    if state in ("present",):
        if volumeByName is not None:
            module.exit_json(changed=False)

        payload = {"name": name, "project": project, "size": size, "volume_type": volume_type}

        response = account_api.post("/volumes", payload)

        if response.ok:
            module.exit_json(changed=True, data=response.json)

        module.fail_json(msg=f"Error creating volume [{response.status_code}: {response.json}]")

    elif state in ("absent",):
        if volumeByName is None:
            module.exit_json(changed=False)

        if module.check_mode:
            module.exit_json(changed=True)

        response = account_api.delete(f"/volumes/{volumeByName['id']}")
        if response.status_code == 204:
            module.exit_json(changed=True, data=response.json)

        module.fail_json(msg=f"Error deleting volume [{response.status_code}: {response.json}]")


def main():
    argument_spec = scaleway_argument_spec()
    argument_spec.update(
        dict(
            state=dict(default="present", choices=["absent", "present"]),
            name=dict(required=True),
            size=dict(type="int"),
            project=dict(),
            organization=dict(),
            volume_type=dict(),
            region=dict(required=True, choices=list(SCALEWAY_LOCATION.keys())),
        )
    )
    module = AnsibleModule(
        argument_spec=argument_spec,
        supports_check_mode=True,
        mutually_exclusive=[
            ("organization", "project"),
        ],
        required_one_of=[
            ("organization", "project"),
        ],
    )

    core(module)


if __name__ == "__main__":
    main()



#!/usr/bin/python

# Copyright (c) 2018, Johannes Brunswicker <johannes.brunswicker@gmail.com>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

DOCUMENTATION = r"""
module: utm_proxy_frontend

author:
  - Johannes Brunswicker (@MatrixCrawler)

short_description: Create, update or destroy reverse_proxy frontend entry in Sophos UTM

description:
  - Create, update or destroy a reverse_proxy frontend entry in Sophos UTM.
  - This module needs to have the REST Ability of the UTM to be activated.
attributes:
  check_mode:
    support: none
  diff_mode:
    support: none

options:
  name:
    type: str
    description:
      - The name of the object that identifies the entry.
    required: true
  add_content_type_header:
    description:
      - Whether to add the content type header or not.
    type: bool
    default: false
  address:
    type: str
    description:
      - The reference name of the network/interface_address object.
    default: REF_DefaultInternalAddress
  allowed_networks:
    type: list
    elements: str
    description:
      - A list of reference names for the allowed networks.
    default: ['REF_NetworkAny']
  certificate:
    type: str
    description:
      - The reference name of the ca/host_key_cert object.
    default: ""
  comment:
    type: str
    description:
      - An optional comment to add to the object.
    default: ""
  disable_compression:
    description:
      - Whether to enable the compression.
    type: bool
    default: false
  domain:
    type: list
    elements: str
    description:
      - A list of domain names for the frontend object.
  exceptions:
    type: list
    elements: str
    description:
      - A list of exception ref names (reverse_proxy/exception).
    default: []
  htmlrewrite:
    description:
      - Whether to enable HTML rewrite or not.
    type: bool
    default: false
  htmlrewrite_cookies:
    description:
      - Whether to enable HTML rewrite cookie or not.
    type: bool
    default: false
  implicitredirect:
    description:
      - Whether to enable implicit redirection or not.
    type: bool
    default: false
  lbmethod:
    type: str
    description:
      - Which loadbalancer method should be used.
    choices:
      - ""
      - bybusyness
      - bytraffic
      - byrequests
    default: bybusyness
  locations:
    type: list
    elements: str
    description:
      - A list of location ref names (reverse_proxy/location).
    default: []
  port:
    type: int
    description:
      - The frontend http port.
    default: 80
  preservehost:
    description:
      - Whether to preserve host header.
    type: bool
    default: false
  profile:
    type: str
    description:
      - The reference string of the reverse_proxy/profile.
    default: ""
  status:
    description:
      - Whether to activate the frontend entry or not.
    type: bool
    default: true
  type:
    type: str
    description:
      - Which protocol should be used.
    choices:
      - http
      - https
    default: http
  xheaders:
    description:
      - Whether to pass the host header or not.
    type: bool
    default: false

extends_documentation_fragment:
  - community.general._utm
  - community.general._attributes
"""

EXAMPLES = r"""
- name: Create utm proxy_frontend
  community.general.utm_proxy_frontend:
    utm_host: sophos.host.name
    utm_token: abcdefghijklmno1234
    name: TestFrontendEntry
    host: REF_OBJECT_STRING
    state: present

- name: Remove utm proxy_frontend
  community.general.utm_proxy_frontend:
    utm_host: sophos.host.name
    utm_token: abcdefghijklmno1234
    name: TestFrontendEntry
    state: absent
"""

RETURN = r"""
result:
  description: The utm object that was created.
  returned: success
  type: complex
  contains:
    _ref:
      description: The reference name of the object.
      type: str
    _locked:
      description: Whether or not the object is currently locked.
      type: bool
    _type:
      description: The type of the object.
      type: str
    name:
      description: The name of the object.
      type: str
    add_content_type_header:
      description: Whether to add the content type header.
      type: bool
    address:
      description: The reference name of the address.
      type: str
    allowed_networks:
      description: List of reference names of networks associated.
      type: list
    certificate:
      description: Reference name of certificate (ca/host_key_cert).
      type: str
    comment:
      description: The comment string.
      type: str
    disable_compression:
      description: State of compression support.
      type: bool
    domain:
      description: List of hostnames.
      type: list
    exceptions:
      description: List of associated proxy exceptions.
      type: list
    htmlrewrite:
      description: State of HTML rewrite.
      type: bool
    htmlrewrite_cookies:
      description: Whether the HTML rewrite cookie is set.
      type: bool
    implicitredirect:
      description: Whether to use implicit redirection.
      type: bool
    lbmethod:
      description: The method of loadbalancer to use.
      type: str
    locations:
      description: The reference names of reverse_proxy/locations associated with the object.
      type: list
    port:
      description: The port of the frontend connection.
      type: int
    preservehost:
      description: Preserve host header.
      type: bool
    profile:
      description: The associated reverse_proxy/profile.
      type: str
    status:
      description: Whether the frontend object is active or not.
      type: bool
    type:
      description: The connection type.
      type: str
    xheaders:
      description: The xheaders state.
      type: bool
"""

from ansible_collections.community.general.plugins.module_utils._utm_utils import UTM, UTMModule


def main():
    endpoint = "reverse_proxy/frontend"
    key_to_check_for_changes = [
        "add_content_type_header",
        "address",
        "allowed_networks",
        "certificate",
        "comment",
        "disable_compression",
        "domain",
        "exceptions",
        "htmlrewrite",
        "htmlrewrite_cookies",
        "implicitredirect",
        "lbmethod",
        "locations",
        "port",
        "preservehost",
        "profile",
        "status",
        "type",
        "xheaders",
    ]
    module = UTMModule(
        argument_spec=dict(
            name=dict(type="str", required=True),
            add_content_type_header=dict(type="bool", default=False),
            address=dict(type="str", default="REF_DefaultInternalAddress"),
            allowed_networks=dict(type="list", elements="str", default=["REF_NetworkAny"]),
            certificate=dict(type="str", default=""),
            comment=dict(type="str", default=""),
            disable_compression=dict(type="bool", default=False),
            domain=dict(type="list", elements="str"),
            exceptions=dict(type="list", elements="str", default=[]),
            htmlrewrite=dict(type="bool", default=False),
            htmlrewrite_cookies=dict(type="bool", default=False),
            implicitredirect=dict(type="bool", default=False),
            lbmethod=dict(type="str", default="bybusyness", choices=["bybusyness", "bytraffic", "byrequests", ""]),
            locations=dict(type="list", elements="str", default=[]),
            port=dict(type="int", default=80),
            preservehost=dict(type="bool", default=False),
            profile=dict(type="str", default=""),
            status=dict(type="bool", default=True),
            type=dict(type="str", default="http", choices=["http", "https"]),
            xheaders=dict(type="bool", default=False),
        )
    )
    try:
        UTM(module, endpoint, key_to_check_for_changes).execute()
    except Exception as e:
        module.fail_json(msg=f"{e}")


if __name__ == "__main__":
    main()



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright (c) 2021, Cisco Systems
# GNU General Public License v3.0+ (see LICENSE or https://www.gnu.org/licenses/gpl-3.0.txt)

DOCUMENTATION = r"""
module: networks_wireless_signal_quality_history_info
short_description: Information module for networks _wireless _signalqualityhistory
description:
  - Information module for Networks Wireless Signalqualityhistory Info.
  - Get all networks _wireless _signalqualityhistory.
  - Return signal quality SNR/RSSI over time for a device or network client.
version_added: '1.0.0'
extends_documentation_fragment:
  - cisco.meraki.module_info
author: Francisco Munoz (@fmunoz)
options:
  headers:
    description: Additional headers.
    type: dict
  networkId:
    description:
      - Information module for Networks Wireless Signalqualityhistory Info.
      - NetworkId path parameter. Network ID.
    type: str
  t0:
    description:
      - Information module for Networks Wireless Signalqualityhistory Info.
      - T0 query parameter. The beginning of the timespan for the data. The maximum
        lookback period is 31 days from today.
    type: str
  t1:
    description:
      - Information module for Networks Wireless Signalqualityhistory Info.
      - T1 query parameter. The end of the timespan for the data. T1 can be a maximum
        of 31 days after t0.
    type: str
  timespan:
    description:
      - Information module for Networks Wireless Signalqualityhistory Info.
      - >
        Timespan query parameter. The timespan for which the information will be fetched.
        If specifying timespan, do not specify parameters t0 and t1. The value must
        be in seconds and be less than or equal to 31 days. The default is 7 days.
    type: float
  resolution:
    description:
      - Information module for Networks Wireless Signalqualityhistory Info.
      - >
        Resolution query parameter. The time resolution in seconds for returned data.
        The valid resolutions are 300, 600, 1200, 3600, 14400, 86400. The default
        is 86400.
    type: int
  autoResolution:
    description:
      - Information module for Networks Wireless Signalqualityhistory Info.
      - >
        AutoResolution query parameter. Automatically select a data resolution based
        on the given timespan; this overrides the value specified by the 'resolution'
        parameter. The default setting is false.
    type: bool
  clientId:
    description:
      - Information module for Networks Wireless Signalqualityhistory Info.
      - ClientId query parameter. Filter results by network client.
    type: str
  deviceSerial:
    description:
      - Information module for Networks Wireless Signalqualityhistory Info.
      - DeviceSerial query parameter. Filter results by device.
    type: str
  apTag:
    description:
      - Information module for Networks Wireless Signalqualityhistory Info.
      - ApTag query parameter. Filter results by AP tag; either clientId or deviceSerial
        must be jointly specified.
    type: str
  band:
    description:
      - Information module for Networks Wireless Signalqualityhistory Info.
      - Band query parameter. Filter results by band (either '2.4', '5' or '6').
    type: str
  ssid:
    description:
      - Information module for Networks Wireless Signalqualityhistory Info.
      - Ssid query parameter. Filter results by SSID number.
    type: int
requirements:
  - meraki >= 2.4.9
  - python >= 3.5
seealso:
  - name: Cisco Meraki documentation for wireless getNetworkWirelessSignalQualityHistory
    description: Complete reference of the getNetworkWirelessSignalQualityHistory
      API.
    link: https://developer.cisco.com/meraki/api-v1/#!get-network-wireless-signal-quality-history
notes:
  - SDK Method used are
    wireless.Wireless.get_network_wireless_signal_quality_history,
  - Paths used are
    get /networks/{networkId}/wireless/signalQualityHistory,
"""

EXAMPLES = r"""
- name: Get all networks _wireless _signalqualityhistory
  cisco.meraki.networks_wireless_signal_quality_history_info:
    meraki_api_key: "{{ meraki_api_key }}"
    meraki_base_url: "{{ meraki_base_url }}"
    meraki_single_request_timeout: "{{ meraki_single_request_timeout }}"
    meraki_certificate_path: "{{ meraki_certificate_path }}"
    meraki_requests_proxy: "{{ meraki_requests_proxy }}"
    meraki_wait_on_rate_limit: "{{ meraki_wait_on_rate_limit }}"
    meraki_nginx_429_retry_wait_time: "{{ meraki_nginx_429_retry_wait_time }}"
    meraki_action_batch_retry_wait_time: "{{ meraki_action_batch_retry_wait_time }}"
    meraki_retry_4xx_error: "{{ meraki_retry_4xx_error }}"
    meraki_retry_4xx_error_wait_time: "{{ meraki_retry_4xx_error_wait_time }}"
    meraki_maximum_retries: "{{ meraki_maximum_retries }}"
    meraki_output_log: "{{ meraki_output_log }}"
    meraki_log_file_prefix: "{{ meraki_log_file_prefix }}"
    meraki_log_path: "{{ meraki_log_path }}"
    meraki_print_console: "{{ meraki_print_console }}"
    meraki_suppress_logging: "{{ meraki_suppress_logging }}"
    meraki_simulate: "{{ meraki_simulate }}"
    meraki_be_geo_id: "{{ meraki_be_geo_id }}"
    meraki_caller: "{{ meraki_caller }}"
    meraki_use_iterator_for_get_pages: "{{ meraki_use_iterator_for_get_pages }}"
    meraki_inherit_logging_config: "{{ meraki_inherit_logging_config }}"
    t0: string
    t1: string
    timespan: 0
    resolution: 0
    autoResolution: true
    clientId: string
    deviceSerial: string
    apTag: string
    band: string
    ssid: 0
    networkId: string
  register: result
"""
RETURN = r"""
meraki_response:
  description: A dictionary or list with the response returned by the Cisco Meraki Python SDK
  returned: always
  type: list
  elements: dict
  sample: >
    [
      {
        "startTs": "string",
        "endTs": "string",
        "snr": 0,
        "rssi": 0
      }
    ]
"""



#!/usr/bin/python

# Copyright (c) 2016, Renato Orgito <orgito@gmail.com>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

DOCUMENTATION = r"""
module: spectrum_device
short_description: Creates/deletes devices in CA Spectrum
description:
  - This module allows you to create and delete devices in CA Spectrum U(https://www.ca.com/us/products/ca-spectrum.html).
  - Tested on CA Spectrum 9.4.2, 10.1.1 and 10.2.1.
author: "Renato Orgito (@orgito)"
extends_documentation_fragment:
  - community.general._attributes
attributes:
  check_mode:
    support: full
  diff_mode:
    support: none
options:
  device:
    type: str
    aliases: [host, name]
    required: true
    description:
      - IP address of the device.
      - If a hostname is given, it is resolved to the IP address.
  community:
    type: str
    description:
      - SNMP community used for device discovery.
      - Required when O(state=present).
    required: true
  landscape:
    type: str
    required: true
    description:
      - Landscape handle of the SpectroServer to which add or remove the device.
  state:
    type: str
    description:
      - On V(present) creates the device when it does not exist.
      - On V(absent) removes the device when it exists.
    choices: ['present', 'absent']
    default: 'present'
  url:
    type: str
    aliases: [oneclick_url]
    required: true
    description:
      - HTTP, HTTPS URL of the Oneclick server in the form V((http|https\)://host.domain[:port]).
  url_username:
    type: str
    aliases: [oneclick_user]
    required: true
    description:
      - Oneclick user name.
  url_password:
    type: str
    aliases: [oneclick_password]
    required: true
    description:
      - Oneclick user password.
  use_proxy:
    description:
      - If V(false), it does not use a proxy, even if one is defined in an environment variable on the target hosts.
    default: true
    type: bool
  validate_certs:
    description:
      - If V(false), SSL certificates are not validated. This should only be used on personally controlled sites using self-signed
        certificates.
    default: true
    type: bool
  agentport:
    type: int
    description:
      - UDP port used for SNMP discovery.
    default: 161
notes:
  - The devices are created inside the I(Universe) container of the specified landscape.
  - All the operations are performed only on the specified landscape.
"""

EXAMPLES = r"""
- name: Add device to CA Spectrum
  local_action:
    module: spectrum_device
    device: '{{ ansible_host }}'
    community: secret
    landscape: '0x100000'
    oneclick_url: http://oneclick.example.com:8080
    oneclick_user: username
    oneclick_password: password
    state: present


- name: Remove device from CA Spectrum
  local_action:
    module: spectrum_device
    device: '{{ ansible_host }}'
    landscape: '{{ landscape_handle }}'
    oneclick_url: http://oneclick.example.com:8080
    oneclick_user: username
    oneclick_password: password
    use_proxy: false
    state: absent
"""

RETURN = r"""
device:
  description: Device data when O(state=present).
  returned: success
  type: dict
  sample:
    {
      "model_handle": "0x1007ab",
      "landscape": "0x100000",
      "address": "10.10.5.1"
    }
"""

import xml.etree.ElementTree as ET
from socket import gaierror, gethostbyname

from ansible.module_utils.basic import AnsibleModule
from ansible.module_utils.urls import fetch_url


def request(resource, xml=None, method=None):
    headers = {"Content-Type": "application/xml", "Accept": "application/xml"}

    url = f"{module.params['oneclick_url']}/spectrum/restful/{resource}"

    response, info = fetch_url(module, url, data=xml, method=method, headers=headers, timeout=45)

    if info["status"] == 401:
        module.fail_json(msg="failed to authenticate to Oneclick server")

    if info["status"] not in (200, 201, 204):
        module.fail_json(msg=info["msg"])

    return response.read()


def post(resource, xml=None):
    return request(resource, xml=xml, method="POST")


def delete(resource):
    return request(resource, xml=None, method="DELETE")


def get_ip():
    try:
        device_ip = gethostbyname(module.params.get("device"))
    except gaierror:
        module.fail_json(msg=f"failed to resolve device ip address for '{module.params.get('device')}'")

    return device_ip


def get_device(device_ip):
    """Query OneClick for the device using the IP Address"""
    resource = "/models"
    landscape_min = f"0x{int(module.params.get('landscape'), 16):x}"
    landscape_max = f"0x{int(module.params.get('landscape'), 16) + 0x100000:x}"

    xml = f"""<?xml version="1.0" encoding="UTF-8"?>
        <rs:model-request throttlesize="5"
        xmlns:rs="http://www.ca.com/spectrum/restful/schema/request"
        xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance"
        xsi:schemaLocation="http://www.ca.com/spectrum/restful/schema/request ../../../xsd/Request.xsd">
            <rs:target-models>
            <rs:models-search>
                <rs:search-criteria xmlns="http://www.ca.com/spectrum/restful/schema/filter">
                    <action-models>
                        <filtered-models>
                            <and>
                                <equals>
                                    <model-type>SearchManager</model-type>
                                </equals>
                                <greater-than>
                                    <attribute id="0x129fa">
                                        <value>{landscape_min}</value>
                                    </attribute>
                                </greater-than>
                                <less-than>
                                    <attribute id="0x129fa">
                                        <value>{landscape_max}</value>
                                    </attribute>
                                </less-than>
                            </and>
                        </filtered-models>
                        <action>FIND_DEV_MODELS_BY_IP</action>
                        <attribute id="AttributeID.NETWORK_ADDRESS">
                            <value>{device_ip}</value>
                        </attribute>
                    </action-models>
                </rs:search-criteria>
            </rs:models-search>
            </rs:target-models>
            <rs:requested-attribute id="0x12d7f" /> <!--Network Address-->
        </rs:model-request>
        """

    result = post(resource, xml=xml)

    root = ET.fromstring(result)

    if root.get("total-models") == "0":
        return None

    namespace = dict(ca="http://www.ca.com/spectrum/restful/schema/response")

    # get the first device
    model = root.find("ca:model-responses", namespace).find("ca:model", namespace)

    if model.get("error"):
        module.fail_json(msg=f"error checking device: {model.get('error')}")

    # get the attributes
    model_handle = model.get("mh")

    model_address = model.find('./*[@id="0x12d7f"]').text

    # derive the landscape handler from the model handler of the device
    model_landscape = f"0x{int(model_handle, 16) // 0x100000 * 0x100000:x}"

    device = dict(model_handle=model_handle, address=model_address, landscape=model_landscape)

    return device


def add_device():
    device_ip = get_ip()
    device = get_device(device_ip)

    if device:
        module.exit_json(changed=False, device=device)

    if module.check_mode:
        device = dict(model_handle=None, address=device_ip, landscape=f"0x{int(module.params.get('landscape'), 16):x}")
        module.exit_json(changed=True, device=device)

    resource = f"model?ipaddress={device_ip}&commstring={module.params.get('community')}"
    resource += f"&landscapeid={module.params.get('landscape')}"

    if module.params.get("agentport", None):
        resource += f"&agentport={module.params.get('agentport', 161)}"

    result = post(resource)
    root = ET.fromstring(result)

    if root.get("error") != "Success":
        module.fail_json(msg=root.get("error-message"))

    namespace = dict(ca="http://www.ca.com/spectrum/restful/schema/response")
    model = root.find("ca:model", namespace)

    model_handle = model.get("mh")
    model_landscape = f"0x{int(model_handle, 16) // 0x100000 * 0x100000:x}"

    device = dict(
        model_handle=model_handle,
        address=device_ip,
        landscape=model_landscape,
    )

    module.exit_json(changed=True, device=device)


def remove_device():
    device_ip = get_ip()
    device = get_device(device_ip)

    if device is None:
        module.exit_json(changed=False)

    if module.check_mode:
        module.exit_json(changed=True)

    resource = f"/model/{device['model_handle']}"
    result = delete(resource)

    root = ET.fromstring(result)

    namespace = dict(ca="http://www.ca.com/spectrum/restful/schema/response")
    error = root.find("ca:error", namespace).text

    if error != "Success":
        error_message = root.find("ca:error-message", namespace).text
        module.fail_json(msg=f"{error} {error_message}")

    module.exit_json(changed=True)


def main():
    global module
    module = AnsibleModule(
        argument_spec=dict(
            device=dict(required=True, aliases=["host", "name"]),
            landscape=dict(required=True),
            state=dict(choices=["present", "absent"], default="present"),
            community=dict(required=True, no_log=True),  # @TODO remove the 'required', given the required_if ?
            agentport=dict(type="int", default=161),
            url=dict(required=True, aliases=["oneclick_url"]),
            url_username=dict(required=True, aliases=["oneclick_user"]),
            url_password=dict(required=True, no_log=True, aliases=["oneclick_password"]),
            use_proxy=dict(type="bool", default=True),
            validate_certs=dict(type="bool", default=True),
        ),
        required_if=[("state", "present", ["community"])],
        supports_check_mode=True,
    )

    if module.params.get("state") == "present":
        add_device()
    else:
        remove_device()


if __name__ == "__main__":
    main()



#!/usr/bin/python
# Copyright Ansible Project
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

DOCUMENTATION = r"""
module: vertica_user
short_description: Adds or removes Vertica database users and assigns roles
description:
  - Adds or removes Vertica database user and, optionally, assigns roles.
  - A user is not removed until all the dependencies have been dropped.
  - In such a situation, if the module tries to remove the user it fails and only remove roles granted to the user.
extends_documentation_fragment:
  - community.general._attributes
attributes:
  check_mode:
    support: full
  diff_mode:
    support: none
options:
  user:
    description:
      - Name of the user to add or remove.
    required: true
    type: str
    aliases: ['name']
  profile:
    description:
      - Sets the user's profile.
    type: str
  resource_pool:
    description:
      - Sets the user's resource pool.
    type: str
  password:
    description:
      - The user's password encrypted by the MD5 algorithm.
      - The password must be generated with the format C("md5" + md5[password + username]), resulting in a total of 35 characters.
        An easy way to do this is by querying the Vertica database with select V('md5'|| md5('<user_password><user_name>'\)).
    type: str
  expired:
    description:
      - Sets the user's password expiration.
    type: bool
  ldap:
    description:
      - Set to V(true) if users are authenticated using LDAP.
      - The user is created with password expired and set to V($ldap$).
    type: bool
  roles:
    description:
      - Comma separated list of roles to assign to the user.
    aliases: ['role']
    type: str
  state:
    description:
      - Whether to create (V(present)), drop (V(absent)), or lock (V(locked)) a user.
    choices: ['present', 'absent', 'locked']
    default: present
    type: str
  db:
    description:
      - Name of the Vertica database.
    type: str
  cluster:
    description:
      - Name of the Vertica cluster.
    default: localhost
    type: str
  port:
    description:
      - Vertica cluster port to connect to.
    default: '5433'
    type: str
  login_user:
    description:
      - The username used to authenticate with.
    default: dbadmin
    type: str
  login_password:
    description:
      - The password used to authenticate with.
    type: str
notes:
  - The default authentication assumes that you are either logging in as or sudo'ing to the C(dbadmin) account on the host.
  - This module uses C(pyodbc), a Python ODBC database adapter. You must ensure that C(unixODBC) and C(pyodbc) is installed
    on the host and properly configured.
  - Configuring C(unixODBC) for Vertica requires C(Driver = /opt/vertica/lib64/libverticaodbc.so) to be added to the C(Vertica)
    section of either C(/etc/odbcinst.ini) or C($HOME/.odbcinst.ini) and both C(ErrorMessagesPath = /opt/vertica/lib64) and
    C(DriverManagerEncoding = UTF-16) to be added to the C(Driver) section of either C(/etc/vertica.ini) or C($HOME/.vertica.ini).
requirements: ['unixODBC', 'pyodbc']
author: "Dariusz Owczarek (@dareko)"
"""

EXAMPLES = r"""
- name: Creating a new vertica user with password
  community.general.vertica_user: name=user_name password=md5<encrypted_password> db=db_name state=present

- name: Creating a new vertica user authenticated via ldap with roles assigned
  community.general.vertica_user: name=user_name ldap=true db=db_name roles=schema_name_ro state=present
"""
import traceback

PYODBC_IMP_ERR = None
try:
    import pyodbc
except ImportError:
    PYODBC_IMP_ERR = traceback.format_exc()
    pyodbc_found = False
else:
    pyodbc_found = True

from ansible.module_utils.basic import AnsibleModule, missing_required_lib


class NotSupportedError(Exception):
    pass


class CannotDropError(Exception):
    pass


# module specific functions


def get_user_facts(cursor, user=""):
    facts = {}
    cursor.execute(
        """
        select u.user_name, u.is_locked, u.lock_time,
        p.password, p.acctexpired as is_expired,
        u.profile_name, u.resource_pool,
        u.all_roles, u.default_roles
        from users u join password_auditor p on p.user_id = u.user_id
        where not u.is_super_user
        and (? = '' or u.user_name ilike ?)
    """,
        user,
        user,
    )
    while True:
        rows = cursor.fetchmany(100)
        if not rows:
            break
        for row in rows:
            user_key = row.user_name.lower()
            facts[user_key] = {
                "name": row.user_name,
                "locked": str(row.is_locked),
                "password": row.password,
                "expired": str(row.is_expired),
                "profile": row.profile_name,
                "resource_pool": row.resource_pool,
                "roles": [],
                "default_roles": [],
            }
            if row.is_locked:
                facts[user_key]["locked_time"] = str(row.lock_time)
            if row.all_roles:
                facts[user_key]["roles"] = row.all_roles.replace(" ", "").split(",")
            if row.default_roles:
                facts[user_key]["default_roles"] = row.default_roles.replace(" ", "").split(",")
    return facts


def update_roles(user_facts, cursor, user, existing_all, existing_default, required):
    del_roles = list(set(existing_all) - set(required))
    if del_roles:
        cursor.execute(f"revoke {','.join(del_roles)} from {user}")
    new_roles = list(set(required) - set(existing_all))
    if new_roles:
        cursor.execute(f"grant {','.join(new_roles)} to {user}")
    if required:
        cursor.execute(f"alter user {user} default role {','.join(required)}")


def check(user_facts, user, profile, resource_pool, locked, password, expired, ldap, roles):
    user_key = user.lower()
    if user_key not in user_facts:
        return False
    if profile and profile != user_facts[user_key]["profile"]:
        return False
    if resource_pool and resource_pool != user_facts[user_key]["resource_pool"]:
        return False
    if locked != (user_facts[user_key]["locked"] == "True"):
        return False
    if password and password != user_facts[user_key]["password"]:
        return False
    if (
        expired is not None
        and expired != (user_facts[user_key]["expired"] == "True")
        or ldap is not None
        and ldap != (user_facts[user_key]["expired"] == "True")
    ):
        return False
    return not (
        roles
        and (
            sorted(roles) != sorted(user_facts[user_key]["roles"])
            or sorted(roles) != sorted(user_facts[user_key]["default_roles"])
        )
    )


def present(user_facts, cursor, user, profile, resource_pool, locked, password, expired, ldap, roles):
    user_key = user.lower()
    if user_key not in user_facts:
        query_fragments = [f"create user {user}"]
        if locked:
            query_fragments.append("account lock")
        if password or ldap:
            if password:
                query_fragments.append(f"identified by '{password}'")
            else:
                query_fragments.append("identified by '$ldap$'")
        if expired or ldap:
            query_fragments.append("password expire")
        if profile:
            query_fragments.append(f"profile {profile}")
        if resource_pool:
            query_fragments.append(f"resource pool {resource_pool}")
        cursor.execute(" ".join(query_fragments))
        if resource_pool and resource_pool != "general":
            cursor.execute(f"grant usage on resource pool {resource_pool} to {user}")
        update_roles(user_facts, cursor, user, [], [], roles)
        user_facts.update(get_user_facts(cursor, user))
        return True
    else:
        changed = False
        query_fragments = [f"alter user {user}"]
        if locked is not None and locked != (user_facts[user_key]["locked"] == "True"):
            if locked:
                state = "lock"
            else:
                state = "unlock"
            query_fragments.append(f"account {state}")
            changed = True
        if password and password != user_facts[user_key]["password"]:
            query_fragments.append(f"identified by '{password}'")
            changed = True
        if ldap:
            if ldap != (user_facts[user_key]["expired"] == "True"):
                query_fragments.append("password expire")
                changed = True
        elif expired is not None and expired != (user_facts[user_key]["expired"] == "True"):
            if expired:
                query_fragments.append("password expire")
                changed = True
            else:
                raise NotSupportedError("Unexpiring user password is not supported.")
        if profile and profile != user_facts[user_key]["profile"]:
            query_fragments.append(f"profile {profile}")
            changed = True
        if resource_pool and resource_pool != user_facts[user_key]["resource_pool"]:
            query_fragments.append(f"resource pool {resource_pool}")
            if user_facts[user_key]["resource_pool"] != "general":
                cursor.execute(f"revoke usage on resource pool {user_facts[user_key]['resource_pool']} from {user}")
            if resource_pool != "general":
                cursor.execute(f"grant usage on resource pool {resource_pool} to {user}")
            changed = True
        if changed:
            cursor.execute(" ".join(query_fragments))
        if roles and (
            sorted(roles) != sorted(user_facts[user_key]["roles"])
            or sorted(roles) != sorted(user_facts[user_key]["default_roles"])
        ):
            update_roles(
                user_facts, cursor, user, user_facts[user_key]["roles"], user_facts[user_key]["default_roles"], roles
            )
            changed = True
        if changed:
            user_facts.update(get_user_facts(cursor, user))
        return changed


def absent(user_facts, cursor, user, roles):
    user_key = user.lower()
    if user_key in user_facts:
        update_roles(user_facts, cursor, user, user_facts[user_key]["roles"], user_facts[user_key]["default_roles"], [])
        try:
            cursor.execute(f"drop user {user_facts[user_key]['name']}")
        except pyodbc.Error as e:
            raise CannotDropError("Dropping user failed due to dependencies.") from e
        del user_facts[user_key]
        return True
    else:
        return False


# module logic


def main():
    module = AnsibleModule(
        argument_spec=dict(
            user=dict(required=True, aliases=["name"]),
            profile=dict(),
            resource_pool=dict(),
            password=dict(no_log=True),
            expired=dict(type="bool"),
            ldap=dict(type="bool"),
            roles=dict(aliases=["role"]),
            state=dict(default="present", choices=["absent", "present", "locked"]),
            db=dict(),
            cluster=dict(default="localhost"),
            port=dict(default="5433"),
            login_user=dict(default="dbadmin"),
            login_password=dict(no_log=True),
        ),
        supports_check_mode=True,
    )

    if not pyodbc_found:
        module.fail_json(msg=missing_required_lib("pyodbc"), exception=PYODBC_IMP_ERR)

    user = module.params["user"]
    profile = module.params["profile"]
    if profile:
        profile = profile.lower()
    resource_pool = module.params["resource_pool"]
    if resource_pool:
        resource_pool = resource_pool.lower()
    password = module.params["password"]
    expired = module.params["expired"]
    ldap = module.params["ldap"]
    roles = []
    if module.params["roles"]:
        roles = module.params["roles"].split(",")
        roles = [_f for _f in roles if _f]
    state = module.params["state"]
    if state == "locked":
        locked = True
    else:
        locked = False
    db = ""
    if module.params["db"]:
        db = module.params["db"]

    changed = False

    try:
        dsn = (
            "Driver=Vertica;"
            f"Server={module.params['cluster']};"
            f"Port={module.params['port']};"
            f"Database={db};"
            f"User={module.params['login_user']};"
            f"Password={module.params['login_password']};"
            f"ConnectionLoadBalance=true"
        )
        db_conn = pyodbc.connect(dsn, autocommit=True)
        cursor = db_conn.cursor()
    except Exception as e:
        module.fail_json(msg=f"Unable to connect to database: {e}.")

    try:
        user_facts = get_user_facts(cursor)
        if module.check_mode:
            changed = not check(user_facts, user, profile, resource_pool, locked, password, expired, ldap, roles)
        elif state == "absent":
            try:
                changed = absent(user_facts, cursor, user, roles)
            except pyodbc.Error as e:
                module.fail_json(msg=f"{e}", exception=traceback.format_exc())
        elif state in ["present", "locked"]:
            try:
                changed = present(
                    user_facts, cursor, user, profile, resource_pool, locked, password, expired, ldap, roles
                )
            except pyodbc.Error as e:
                module.fail_json(msg=f"{e}", exception=traceback.format_exc())
    except NotSupportedError as e:
        module.fail_json(msg=f"{e}", ansible_facts={"vertica_users": user_facts})
    except CannotDropError as e:
        module.fail_json(msg=f"{e}", ansible_facts={"vertica_users": user_facts})
    except SystemExit:
        raise
    except Exception as e:
        module.fail_json(msg=f"{e}", exception=traceback.format_exc())

    module.exit_json(changed=changed, user=user, ansible_facts={"vertica_users": user_facts})


if __name__ == "__main__":
    main()



#!/usr/bin/python
# Copyright (c) 2021, Florian Dambrine <android.florian@gmail.com>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

DOCUMENTATION = r"""
module: pritunl_org_info
author: Florian Dambrine (@Lowess)
version_added: 2.5.0
short_description: List Pritunl Organizations using the Pritunl API
description:
  - A module to list Pritunl organizations using the Pritunl API.
extends_documentation_fragment:
  - community.general._pritunl
  - community.general._attributes
  - community.general._attributes.info_module
options:
  organization:
    type: str
    aliases:
      - org
    default: null
    description:
      - Name of the Pritunl organization to search for. If none provided, the module returns all Pritunl organizations.
"""

EXAMPLES = r"""
- name: List all existing Pritunl organizations
  community.general.pritunl_org_info:

- name: Search for an organization named MyOrg
  community.general.pritunl_user_info:
    organization: MyOrg
"""

RETURN = r"""
organizations:
  description: List of Pritunl organizations.
  returned: success
  type: list
  elements: dict
  sample:
    [
      {
        "auth_api": false,
        "name": "FooOrg",
        "auth_token": null,
        "user_count": 0,
        "auth_secret": null,
        "id": "csftwlu6uhralzi2dpmhekz3"
      },
      {
        "auth_api": false,
        "name": "MyOrg",
        "auth_token": null,
        "user_count": 3,
        "auth_secret": null,
        "id": "58070daee63f3b2e6e472c36"
      },
      {
        "auth_api": false,
        "name": "BarOrg",
        "auth_token": null,
        "user_count": 0,
        "auth_secret": null,
        "id": "v1sncsxxybnsylc8gpqg85pg"
      }
    ]
"""

from ansible.module_utils.basic import AnsibleModule
from ansible.module_utils.common.dict_transformations import dict_merge

from ansible_collections.community.general.plugins.module_utils._pritunl_api import (
    PritunlException,
    get_pritunl_settings,
    list_pritunl_organizations,
    pritunl_argument_spec,
)


def get_pritunl_organizations(module):
    org_name = module.params.get("organization")

    organizations = []

    organizations = list_pritunl_organizations(
        **dict_merge(
            get_pritunl_settings(module),
            {"filters": {"name": org_name} if org_name else None},
        )
    )

    if org_name and len(organizations) == 0:
        # When an org_name is provided but no organization match return an error
        module.fail_json(msg=f"Organization '{org_name}' does not exist")

    result = {}
    result["changed"] = False
    result["organizations"] = organizations

    module.exit_json(**result)


def main():
    argument_spec = pritunl_argument_spec()

    argument_spec.update(dict(organization=dict(type="str", aliases=["org"])))

    module = AnsibleModule(argument_spec=argument_spec, supports_check_mode=True)

    try:
        get_pritunl_organizations(module)
    except PritunlException as e:
        module.fail_json(msg=f"{e}")


if __name__ == "__main__":
    main()



#!/usr/bin/python
#
# Copyright (c) 2018, Yanis Guenane <yanis+ansible@guenane.org>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

DOCUMENTATION = r"""
module: scaleway_image_info
short_description: Gather information about the Scaleway images available
description:
  - Gather information about the Scaleway images available.
author:
  - "Yanis Guenane (@Spredzy)"
  - "Remy Leone (@remyleone)"
extends_documentation_fragment:
  - community.general._scaleway
  - community.general._attributes
  - community.general._scaleway.actiongroup_scaleway
  - community.general._attributes.info_module

attributes:
  action_group:
    version_added: 11.3.0

options:
  region:
    type: str
    description:
      - Scaleway compute zone.
    required: true
    choices:
      - ams1
      - EMEA-NL-EVS
      - ams2
      - ams3
      - par1
      - EMEA-FR-PAR1
      - par2
      - EMEA-FR-PAR2
      - par3
      - waw1
      - EMEA-PL-WAW1
      - waw2
      - waw3
"""

EXAMPLES = r"""
- name: Gather Scaleway images information
  community.general.scaleway_image_info:
    region: par1
  register: result

- ansible.builtin.debug:
    msg: "{{ result.scaleway_image_info }}"
"""

RETURN = r"""
scaleway_image_info:
  description:
    - Response from Scaleway API.
    - 'For more details please refer to: U(https://developers.scaleway.com/en/products/instance/api/).'
  returned: success
  type: list
  elements: dict
  sample:
    [
      {
        "arch": "x86_64",
        "creation_date": "2018-07-17T16:18:49.276456+00:00",
        "default_bootscript": {
          "architecture": "x86_64",
          "bootcmdargs": "LINUX_COMMON scaleway boot=local nbd.max_part=16",
          "default": false,
          "dtb": "",
          "id": "15fbd2f7-a0f9-412b-8502-6a44da8d98b8",
          "initrd": "http://169.254.42.24/initrd/initrd-Linux-x86_64-v3.14.5.gz",
          "kernel": "http://169.254.42.24/kernel/x86_64-mainline-lts-4.9-4.9.93-rev1/vmlinuz-4.9.93",
          "organization": "11111111-1111-4111-8111-111111111111",
          "public": true,
          "title": "x86_64 mainline 4.9.93 rev1"
        },
        "extra_volumes": [],
        "from_server": null,
        "id": "00ae4a88-3252-4eda-9feb-5f6b56bf5ef0",
        "modification_date": "2018-07-17T16:42:06.319315+00:00",
        "name": "Debian Stretch",
        "organization": "51b656e3-4865-41e8-adbc-0c45bdd780db",
        "public": true,
        "root_volume": {
          "id": "da32dfbb-c5ff-476d-ae2d-c297dd09b7dd",
          "name": "snapshot-2a7229dc-d431-4dc5-b66e-95db08b773af-2018-07-17_16:18",
          "size": 25000000000,
          "volume_type": "l_ssd"
        },
        "state": "available"
      }
    ]
"""

from ansible.module_utils.basic import AnsibleModule

from ansible_collections.community.general.plugins.module_utils._scaleway import (
    SCALEWAY_LOCATION,
    Scaleway,
    ScalewayException,
    scaleway_argument_spec,
)


class ScalewayImageInfo(Scaleway):
    def __init__(self, module):
        super().__init__(module)
        self.name = "images"

        region = module.params["region"]
        self.module.params["api_url"] = SCALEWAY_LOCATION[region]["api_endpoint"]


def main():
    argument_spec = scaleway_argument_spec()
    argument_spec.update(
        dict(
            region=dict(required=True, choices=list(SCALEWAY_LOCATION.keys())),
        )
    )
    module = AnsibleModule(
        argument_spec=argument_spec,
        supports_check_mode=True,
    )

    try:
        module.exit_json(scaleway_image_info=ScalewayImageInfo(module).get_resources())
    except ScalewayException as exc:
        module.fail_json(msg=exc.message)


if __name__ == "__main__":
    main()



#!/usr/bin/python

# (c) 2017 David Gunter <david.gunter@tivix.com>
# Copyright (c) 2017 Chris Hoffman <christopher.hoffman@gmail.com>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later


from __future__ import annotations

DOCUMENTATION = r"""
module: yarn
short_description: Manage Node.js packages with Yarn
description:
  - Manage Node.js packages with the Yarn package manager U(https://yarnpkg.com/).
  - Note that at the moment, this module B(only works with Yarn Classic).
author:
  - "David Gunter (@verkaufer)"
  - "Chris Hoffman (@chrishoffman), creator of NPM Ansible module)"
extends_documentation_fragment:
  - community.general._attributes
attributes:
  check_mode:
    support: full
  diff_mode:
    support: none
options:
  name:
    type: str
    description:
      - The name of a Node.js library to install.
      - If omitted all packages in package.json are installed.
      - To globally install from local Node.js library. Prepend C(file:) to the path of the Node.js library.
  path:
    type: path
    description:
      - The base path where Node.js installs libraries.
      - This is where the C(node_modules) folder lives.
  version:
    type: str
    description:
      - The version of the library to be installed.
      - Must be in semver format. If "latest" is desired, use O(state) arg instead.
  global:
    description:
      - Install the Node.js library globally.
    default: false
    type: bool
  executable:
    type: path
    description:
      - The executable location for yarn.
  ignore_scripts:
    description:
      - Use the C(--ignore-scripts) flag when installing.
    type: bool
    default: false
  production:
    description:
      - Install dependencies in production mode.
      - C(yarn) ignores any dependencies under devDependencies in C(package.json).
    type: bool
    default: false
  registry:
    type: str
    description:
      - The registry to install modules from.
  state:
    type: str
    description:
      - Installation state of the named Node.js library.
      - If V(absent) is selected, a O(name) option must be provided.
    default: present
    choices: ["present", "absent", "latest"]
requirements:
  - Yarn Classic installed in bin path (typically C(/usr/local/bin))
"""

EXAMPLES = r"""
- name: Install "imagemin" Node.js package.
  community.general.yarn:
    name: imagemin
    path: /app/location

- name: Install "imagemin" Node.js package on version 5.3.1
  community.general.yarn:
    name: imagemin
    version: '5.3.1'
    path: /app/location

- name: Install "imagemin" Node.js package globally.
  community.general.yarn:
    name: imagemin
    global: true

- name: Remove the globally-installed package "imagemin".
  community.general.yarn:
    name: imagemin
    global: true
    state: absent

- name: Install "imagemin" Node.js package from custom registry.
  community.general.yarn:
    name: imagemin
    registry: 'http://registry.mysite.com'

- name: Install packages based on package.json.
  community.general.yarn:
    path: /app/location

- name: Update all packages in package.json to their latest version.
  community.general.yarn:
    path: /app/location
    state: latest
"""

RETURN = r"""
out:
  description: Output generated from Yarn.
  returned: always
  type: str
  sample: "yarn add v0.16.1[1/4] Resolving packages...[2/4] Fetching packages...[3/4] Linking dependencies...[4/4] Building
    fresh packages...success Saved lockfile.success Saved 1 new dependency..left-pad@1.1.3 Done in 0.59s."
"""

import json
import os

from ansible.module_utils.basic import AnsibleModule


class Yarn:
    def __init__(self, module, **kwargs):
        self.module = module
        self.globally = kwargs["globally"]
        self.name = kwargs["name"]
        self.version = kwargs["version"]
        self.path = kwargs["path"]
        self.registry = kwargs["registry"]
        self.production = kwargs["production"]
        self.ignore_scripts = kwargs["ignore_scripts"]
        self.executable = kwargs["executable"]

        # Specify a version of package if version arg passed in
        self.name_version = None

        if kwargs["version"] and self.name is not None:
            self.name_version = f"{self.name}@{self.version!s}"
        elif self.name is not None:
            self.name_version = self.name

    def _exec(self, args, run_in_check_mode=False, check_rc=True, unsupported_with_global=False):
        if not self.module.check_mode or (self.module.check_mode and run_in_check_mode):
            with_global_arg = self.globally and not unsupported_with_global

            if with_global_arg:
                # Yarn global arg is inserted before the command (e.g. `yarn global {some-command}`)
                args.insert(0, "global")

            cmd = self.executable + args

            if self.production:
                cmd.append("--production")
            if self.ignore_scripts:
                cmd.append("--ignore-scripts")
            if self.registry:
                cmd.append("--registry")
                cmd.append(self.registry)

            # If path is specified, cd into that path and run the command.
            cwd = None
            if self.path and not with_global_arg:
                if not os.path.exists(self.path):
                    # Module will make directory if not exists.
                    os.makedirs(self.path)
                if not os.path.isdir(self.path):
                    self.module.fail_json(msg=f"Path provided {self.path} is not a directory")
                cwd = self.path

                if not os.path.isfile(os.path.join(self.path, "package.json")):
                    self.module.fail_json(msg="Package.json does not exist in provided path.")

            rc, out, err = self.module.run_command(cmd, check_rc=check_rc, cwd=cwd)
            return out, err

        return None, None

    def _process_yarn_error(self, err):
        try:
            # We need to filter for errors, since Yarn warnings are included in stderr.
            # Non-JSON lines (e.g. Node.js runtime warnings) are surfaced via module.warn()
            # rather than treated as errors, since their meaning is unknown.
            for line in err.splitlines():
                if not line.startswith("{"):
                    self.module.warn(f"yarn stderr: {line}")
                    continue
                if json.loads(line)["type"] == "error":
                    self.module.fail_json(msg=err)
        except Exception:
            self.module.fail_json(msg=f"Unexpected stderr output from Yarn: {err}", stderr=err)

    def list(self):
        cmd = ["list", "--depth=0", "--json"]

        installed = list()
        missing = list()

        if not os.path.isfile(os.path.join(self.path, "yarn.lock")):
            missing.append(self.name)
            return installed, missing

        # `yarn global list` should be treated as "unsupported with global" even though it exists,
        # because it only only lists binaries, but `yarn global add` can install libraries too.
        result, error = self._exec(cmd, run_in_check_mode=True, check_rc=False, unsupported_with_global=True)

        self._process_yarn_error(error)

        for json_line in result.strip().split("\n"):
            data = json.loads(json_line)
            if data["type"] == "tree":
                dependencies = data["data"]["trees"]

                for dep in dependencies:
                    name, version = dep["name"].rsplit("@", 1)
                    installed.append(name)

        if self.name not in installed:
            missing.append(self.name)

        return installed, missing

    def install(self):
        if self.name_version:
            # Yarn has a separate command for installing packages by name...
            return self._exec(["add", self.name_version])
        # And one for installing all packages in package.json
        return self._exec(["install", "--non-interactive"])

    def update(self):
        return self._exec(["upgrade", "--latest"])

    def uninstall(self):
        return self._exec(["remove", self.name])

    def list_outdated(self):
        outdated = list()

        if not os.path.isfile(os.path.join(self.path, "yarn.lock")):
            return outdated

        cmd_result, err = self._exec(["outdated", "--json"], True, False, unsupported_with_global=True)

        # the package.json in the global dir is missing a license field, so warnings are expected on stderr
        self._process_yarn_error(err)

        if not cmd_result:
            return outdated

        outdated_packages_data = cmd_result.splitlines()[1]

        data = json.loads(outdated_packages_data)

        try:
            outdated_dependencies = data["data"]["body"]
        except KeyError:
            return outdated

        for dep in outdated_dependencies:
            # Outdated dependencies returned as a list of lists, where
            # item at index 0 is the name of the dependency
            outdated.append(dep[0])
        return outdated


def main():
    arg_spec = dict(
        name=dict(),
        path=dict(type="path"),
        version=dict(),
        production=dict(default=False, type="bool"),
        executable=dict(type="path"),
        registry=dict(),
        state=dict(default="present", choices=["present", "absent", "latest"]),
        ignore_scripts=dict(default=False, type="bool"),
    )
    arg_spec["global"] = dict(default=False, type="bool")
    module = AnsibleModule(argument_spec=arg_spec, supports_check_mode=True)
    module.run_command_environ_update = {"LANGUAGE": "C", "LC_ALL": "C"}

    name = module.params["name"]
    path = module.params["path"]
    version = module.params["version"]
    globally = module.params["global"]
    production = module.params["production"]
    registry = module.params["registry"]
    state = module.params["state"]
    ignore_scripts = module.params["ignore_scripts"]

    # When installing globally, users should not be able to define a path for installation.
    # Require a path if global is False, though!
    if path is None and globally is False:
        module.fail_json(msg="Path must be specified when not using global arg")
    elif path and globally is True:
        module.fail_json(msg="Cannot specify path if doing global installation")

    if state == "absent" and not name:
        module.fail_json(msg="Package must be explicitly named when uninstalling.")
    if state == "latest":
        version = "latest"

    if module.params["executable"]:
        executable = module.params["executable"].split(" ")
    else:
        executable = [module.get_bin_path("yarn", True)]

    # When installing globally, use the defined path for global node_modules
    if globally:
        _rc, out, _err = module.run_command(executable + ["global", "dir"], check_rc=True)
        path = out.strip()

    yarn = Yarn(
        module,
        name=name,
        path=path,
        version=version,
        globally=globally,
        production=production,
        executable=executable,
        registry=registry,
        ignore_scripts=ignore_scripts,
    )

    changed = False
    out = ""
    err = ""
    if state == "present":
        if not name:
            changed = True
            out, err = yarn.install()
        else:
            installed, missing = yarn.list()
            if len(missing):
                changed = True
                out, err = yarn.install()

    elif state == "latest":
        if not name:
            changed = True
            out, err = yarn.install()
        else:
            installed, missing = yarn.list()
            outdated = yarn.list_outdated()
            if len(missing):
                changed = True
                out, err = yarn.install()
            if len(outdated):
                changed = True
                out, err = yarn.update()
    else:
        # state == absent
        installed, missing = yarn.list()
        if name in installed:
            changed = True
            out, err = yarn.uninstall()

    module.exit_json(changed=changed, out=out, err=err)


if __name__ == "__main__":
    main()
